;;;-*- Mode: Lisp; Encoding: (win32:code-page :id 1251) -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Low-level generics and methods

(in-package :ystok.grid)

;;; Methods on basic-grid

(defmethod changed ((grid basic-grid))
  (or (iplusp (allow-additions grid))
      (iplusp (allow-deletions grid))
      (iplusp (allow-edits grid))))

(defmethod (setf changed) (value (grid basic-grid))
  (unless value
    (when (allow-additions grid) (setf (slot-value grid 'allow-additions) 0))
    (when (allow-edits grid)     (setf (slot-value grid 'allow-edits) 0))
    (when (allow-deletions grid) (setf (slot-value grid 'allow-deletions) 0))))

(defun collect-grids (layout &optional test)
 ;;; Args: test On the contrary to capi:map-pane-descendant-children,
  ;;            is only called for children of type basic-grid.
  (unless (typep layout 'basic-grid)		; as grid itself is layout, do not recurse
    (let ((result ()))
      (capi:map-pane-descendant-children layout
        (lambda (child) (when (and (typep child 'basic-grid)
                                   (imp test (funcall test child)))
                          (push child result))))
      result)))

;;; Errors

(defmethod ywi:signal-not-nullable-error ((grid ygrid)
                                          &optional (column (current-column grid)))
 ;;; Args: column Instance of basic-column or title string.
  (error 'conditions::simple-parse-error
         :format-control #L"Column ~s is not nullable - cell value is required."
         :format-arguments (list (ywi:object-title column))))

(defun beep-no-edits (grid)
  (beep (list #L"Record edits are not allowed within grid~@[ '~a'~]."
              (object-title grid))
        grid))

(defun beep-column-read-only (column &optional grid)
  (beep (list #L"Column ~s within grid~@[ '~a'~] is read-only."
              (object-title column) (object-title grid))
        grid))

;;; Updating methods and callbacks

(defun cell-on-action (grid &optional record column x)
 ;;; Helper invoking cell-action-callback if any exists.
  ;; Args: x  If called from a mouse double-click, a coordinate relative to the cell
  ;;          left border.
  ;; Value: True if the action callback has been called,
  ;;        False if none has been called.
  (when-let* ((column (or column (%current-column grid)))
              (cell-action-callback (slot-value column 'cell-action-callback))
              (record (or record (current-record grid))))
    (funcall cell-action-callback grid record column
             (if x
                 (let ((size (compute-natural-size grid column)))
                   (if (iplusp size) (/ x size) nil))
                 nil))
    t))

(defgeneric cell-on-cancel (grid column)
 ;;; Can invoke revert-cell
 (:method (grid column)						; for exotic needs
  (if-bind (function (capi:capi-object-property grid :cancel-function))
    (funcall function grid)
    (beep))))

(defgeneric cell-on-character (grid column character)
 (:documentation "Invoked on typing a character when the pinboard has focus.
  The COLUMN is guaranteed to be enabled, but may be read-only.")
 (:method (grid column character)				; invoked also if corresp.
  (beep))							; accelerator is undefined

 (:method ((grid ygrid) (column column) (character (eql #\Control-\Z)))	; = Ctrl+Shift+Z
  (revert-cell grid (current-record grid) column))

 (:method ((grid ygrid) (column column) (character (eql #\Control-\Space)))
  (complete-cell grid (current-record grid) column))
 ;(:method :after (grid column (character (eql #\Return)))
 ; (cell-on-action grid (%current-record grid) column))
)

(defgeneric cell-on-click (grid row column x y)
 (:documentation "Invoked on mouse click.
 Arguments:
  COLUMN is guaranteed to be enabled
  X,Y are coordinates within the cell rectangle.
 Value: True - needed for popup-column-mixin.")
 (:method (grid row column x y)
  t)
 (:method :around ((grid ygrid) row column x y)
  (let ((col (column-number column)))
    (when (move-cell grid row col t t)
          ;(or (and (eql (current-row grid) row) (eql (current-col grid) col))
      (call-next-method)))) )

(defgeneric cell-on-enter (grid record column)
 ;;; Initialize and enables an editing pane (by assign overlay) if any.
  ;; Q: Are methods allowed to switch input focus to the pane?
  ;; NB1: Should also set %old-value%.
  ;; NB2: As this function can be called before %set-edit-mode on the record,
  ;;      it can actually read from the original record-data object, not from a dirty copy.
 (:method (grid record column))

 (:method :after ((grid ygrid) (record record) column)
  ;; Q: Call (help :pointer-documentation-enter ...) here rather than from move-cell?
  (%invalidate-cell grid (record-row-number record) (column-number column)
                    :redraw-now t)) )

(defgeneric cell-on-exit (grid record column)
 ;;; current-row and -col has not updated yet.
  ;; Value: True = success, false = failure.
 (:method (grid record column)
  (save-cell grid record column))		; should work even if record is NIL

 (:method :around (grid (record record) column)
  (multiple-value-bind (result condition) (call-next-method) ; e.g. <t :ignore>
    (when result
      (with-slots (%old-value%) grid
        (setf %old-value% *unbound-value*)))
        ;(nilf (current-displayer pinboard) (current-subscript pinboard))
    (values result condition)))

 (:method :after (grid record column)
  (declare (ignore record column))
  (help grid :pointer-documentation-leave nil)))

(defgeneric cell-on-update (grid record column &optional recursive)
 ;;; Called after data-writer has succeeded; its value is usually thrown out.
 (:method (grid record column &optional recursive)
  (declare (ignore grid record column recursive)))
 
 (:method ((grid ygrid) (record record) (column column) &optional recursive)
  (when-let (cell-update-callback (slot-value grid 'cell-update-callback))
    (funcall cell-update-callback grid record column recursive))) )

(defgeneric complete-cell-p (grid record column)
 (:method ((grid ygrid) (record null) column)
  nil)
 (:method ((grid ygrid) (record record) (column column))
  (column-completion-function column))
 (:method :around ((grid ygrid) (record record) (column column))
  (when (and (allow-edits grid) (enabled column) (not (column-read-only grid column)))
    (call-next-method))) )

(defgeneric complete-cell (grid record column)
 (:method (grid record column)
  (declare (ignore record))
  ;(message-area-format grid
  ;                     #L"Completion function is undefined for pane or column~@[ ~s~]."
  ;                     (column-title column))
  (beep (format nil #L"Completion function is undefined for pane or column~@[ ~s~]."
                (ywi:object-title column))
        grid)))

(defmethod ywi:pane-complete-p ((grid-component grid-component))
  (let ((grid (capi:element-interface grid-component)))
    (complete-cell-p grid (current-record grid) (current-column grid))))

(defmethod ywi:pane-complete ((grid-component grid-component))
  (let ((grid (capi:element-interface grid-component)))
    (complete-cell grid (current-record grid) (current-column grid))))

(defgeneric grid-confirm-destroy (grid &optional change)
 ;;; Function being set up as the interface confirm-destroy-function.
  ;; Args: change  interface-confirm-destroy
  ;;		    	If called by interface-confirm-destroy or similar.
  ;;		   NIL  If we should clear the changed flag of the grid.
  ;;      		Passed by Ystok-Tournament: <grid-on-activate null>
  ;;			Ystok-Portal: save-interface-object, <operate-on-interface :close>
  ;; NB: confirm-destroy-function gets called automatically only for top-level interfaces.
 (:method-combination and :most-specific-last)

 (:method and ((layout CAPI::PANE-WITH-CHILDREN) &optional (change *unbound-value*))
  ;; Args: layout Instance of interface or layout
  (loop for grid in (collect-grids layout)
        always (grid-confirm-destroy grid change)))

 (:method and ((grid ygrid) &optional (change *unbound-value*))
  (when (or (>= (current-row grid) (nrows grid))
            (and (or (zerop (length (columns grid)))
                     (cell-on-exit grid (%current-record grid) (%current-column grid)))
                 (record-on-exit grid (%current-record grid))))	;:redraw-now nil
    (unless change
      (setf (changed grid) nil))
    t)) )

(defgeneric grid-on-destroy (grid)
 (:method ((face capi:interface))
  (mapc 'grid-on-destroy (collect-grids face)))

 (:method ((grid ygrid))
  (unload-grid grid (name grid))
  (let* ((pinboard (pinboard grid))
         (buffer (buffer pinboard)))
    (when buffer 
      (when (editor:buffer-flag buffer)
        (editor::kill-buffer-no-confirm buffer)) ;(editor:kill-buffer-command)
      (nilf (slot-value pinboard 'buffer))))) )

(defgeneric operate-on-cell (grid record column operation)
 (:documentation "Perform operation of any kind, except for :undo.")
 (:method (grid record column operation)
  (beep-impossible grid operation :name (ywi:object-title column))))

(defgeneric operate-on-cell-p (grid record column operation)
 (:documentation "Predicate returning true if an operation is possible.
Called from operate-on-grid-p.")
  ;; NB: An around method verifying (enabled column) would fit all except :zoom.
  ;; NB: For :undo, the editing panes provide neccessary functionality themselves.
  ;;     Do we need it for popup-column subclasses?
 (:method (grid record column operation)
  nil))

(defgeneric save-cell (grid record column &key value redraw-now &allow-other-keys)
 ;;; Values:
  ;; (T NIL)	  Has changed and was saved successfully,
  ;; (T T)	  Has not changed and no need to save.
  ;; (T :ignore)  Failed but the change is considered OK, so the caller is allowed
  ;;		  to save the whole record.
  ;; (NIL :abort) Failure
 (:method (grid record column &key value redraw-now)
  (declare (ignore grid record column value redraw-now))
  t))

;;; Helpers for operate-on-cell to act on the entire cell contents for edit/select

(defgeneric cell-clip (grid row-or-record column)
 ;;; Values: 1) String or NIL if cell value is NIL or cannot be represented.
  ;;         2) Plist (:format1 value1 ...)
 (:method ((grid ygrid) (record record) column)
  (declare (ignore grid record column format))
  nil))	;+null-string+

(defgeneric set-cell-clip (grid row-or-record column format value &key redraw-now)
 ;;; Value: True = success
 (:method ((grid ygrid) (record record) column format value &key redraw-now)
  (declare (ignore grid record column format value redraw-now))
  nil)

 (:method ((grid ygrid) (record record) (column column) (format (eql :value)) value
           &key (redraw-now :header))
  (save-cell grid record column :value value :redraw-now redraw-now)) )

(defgeneric cell-clip-formats (column)
 ;;; List of preferable clipboard formats to set cell-value
 (:method (column)
  '(:value)))	; or assume :value to be always the best?


(defgeneric grid-confirm-delete (grid object &optional interactive)
 ;;; Args: object ::= :rows | :columns | :both
  ;;       interactive
  ;;	   ::= t	Deletion of one object, usually the specialization should ask.
  ;;         | <number>	When object eql :rows, the number of selected records
  ;;	     | NIL	The specialization should not ask unless something would get wrong.
  ;;
 (:documentation "The function verifying the possibility of deletion and/or asking.")

 (:method ((grid basic-grid) row-or-record &optional (interactive t))
  (or ;(getf (record-plist record) :new)
      (not interactive) 
      (not *confirm-deletions*)
      (beep-box #L"Do you really want to delete the current record~@[~%from '~a' ~]?"
                :format-arguments (list (object-title grid))
                :owner (capi:top-level-interface grid)
                :buttons '(:ok :cancel) :default-button :cancel)))

 (:method ((grid basic-grid) (object (eql :rows)) &optional (interactive t))
  (or (not interactive)
      (not *confirm-deletions*)
      (beep-box #L"Do you really want to delete selected records~@[~%from '~a' ~]?"
                :format-arguments (list (object-title grid))
                :owner (capi:top-level-interface grid)
                :buttons '(:ok :cancel) :default-button :cancel)))

 (:method :around (grid object &optional (interactive t))
  (cond ((allow-deletions grid)
         (call-next-method))
        (interactive
         (beep (list #L"Record deletions are not allowed within grid~@[ '~a'~]."
                     (object-title grid))
               grid)))) )

(defgeneric record-on-delete (grid row-or-record)
 (:documentation "Invoked after the application has deleted record content successfully.
 The default or overriding method should invoke delete-callback and throw out its result.")
 (:method (grid row-or-record)
  ;(cond ((getf (record-plist record) :new))
  ;      ((grid-confirm-delete grid record interactive)
  (let ((allow-deletions (incf (slot-value grid 'allow-deletions))))
    (declare (integer allow-additions)) ;(if (numberp allow-deletions) allow-deletions 0)
    ;(when-let (delete-callback (slot-value grid 'delete-callback))	; DI 2007-Apr-17
    ;  (funcall delete-callback grid record))
    ;; If this is the very first deletion and no others operations have been invoked yet
    (when (and (= allow-deletions 1)
               (not (iplusp (allow-additions grid)))
               (not (iplusp (allow-edits grid))))
      (when-let (change-callback (change-callback grid))
        (funcall change-callback (capi:top-level-interface grid) grid))))
  t))

(defgeneric record-on-change (interface grid row-or-record)
 ;;; Hook "Start editing record" called when the very first cell was just changed.
  ;; "Being edited" flag is actually set only in the after method, so any primary method
  ;; sees the unchanged record.
  ;; A specialization may do one of:
  ;; - create a backup copy of the record-data object,
  ;; - clone it into a dirty temporal copy that would replace the original during edit,
  ;; - check whether the record is not locked by another user.
  ;; Value: Ignored.
  ;;        The primary or before method should signal an error if something went wrong.
 (:method (interface grid row-or-record)
  (declare (ignore interface grid row-or-record)))

 (:method :before (interface (grid ygrid) (record record))
  (declare (ignore interface))
  (when (record-auto-p grid record)
    ;; The user fills in top left cell via ywi:commit -> write-displayer-data
    (record-on-enter grid record)))		; -> %fill-auto-record and current-callback

 (:method :after (interface (grid basic-grid) row-or-record)
  (declare (ignore interface row-or-record))
  (setf (capi:capi-object-property grid :edit) t)) )

(defgeneric record-on-enter (grid record)
 (:method ((grid ygrid) (record record))
  (when (if (record-auto-p grid record)
            (%fill-auto-record grid record)
            (progn (update-navigation-toolbar grid) t))
    (when-let (current-callback (slot-value grid 'current-callback))
      (funcall current-callback grid record)))) )

(defun %record-on-current (grid &optional old (new (current-record grid)))
 ;;; Helper used to react on changing current record in various functions.
  ;; Args: old  Previous current record; omit or specify NIL to force
  (when (and new (neq new old))				; current has moved
    (if (record-auto-p grid new)
        (when-let (current-callback (slot-value grid 'current-callback))
          (funcall current-callback grid new))
        (record-on-enter grid new))))

(defgeneric record-on-exit (grid record &key redraw-now)
 ;;; Called when focus is about to be moved to another record or the grid
  ;; is about to be destroyed.
  ;; The system-supplied method tries to save the record if it has changed.
  ;; Value: True = succeess, false = failure
 (:method (grid record &key redraw-now)
  (declare (ignore grid record redraw-now))
  t)
 (:method (grid (record record) &key (redraw-now t))
  (save-record grid :record record :redraw-now redraw-now :cell nil)) )

(defgeneric record-on-update (grid row-or-record)
 (:documentation "Returns
 True  if the application has updated record content successfully
       (only checked by save-record and its callers, e.g. record-on-exit);
 False or signals on failure.
On success, the effective method should invoke update-callback and
throw out its result.
Both the method and update-callback can observe :changed and :new/:deferred
properties in record-plist to tell insert from update.
The method itself, not update-callback, should remove these properties
(or do something else) only at the end.")
  ;; GOOD: The update-callback observes the record properties not modified.
  ;; BAD: If an :after method modified the record data object additionally,
  ;;      the update-callback would not observe those modifications.
 (:method ((grid ygrid) (record record))
  (let* ((new (getf (record-plist record) :new))
         (count (if new
                    (incf (slot-value grid 'allow-additions))
                    (incf (slot-value grid 'allow-edits)))))
    (declare (integer count))
    (%set-edit-mode grid record nil :redraw-now nil)
    (when-let (update-callback (update-callback grid))
      (funcall update-callback grid record))
    (remf (record-plist record) :changed)
    (remf (record-plist record) :new)
    ;; If this is the very first update and no others operations have been invoked yet
    (when (and (= count 1)
               (if new
                   (and (not (iplusp (allow-edits grid)))
                        (not (iplusp (allow-deletions grid))))
                   (and (not (iplusp (allow-additions grid)))
                        (not (iplusp (allow-deletions grid))))))
      (when-let (change-callback (change-callback grid))
        (funcall change-callback (capi:top-level-interface grid) grid))))
  t))

(defun %after-revert-cell (grid record column &optional exit)
 ;;; Provide record-level clean up if needed
  ;; Args: exit  If true, we are also exiting the cell's overlay or there is no overlay.
  ;; Called by: Methods on revert-cell and also on undo-cell.
  (with-slots (%old-value%) grid
    (if-bind (changed (changed record))
      (when-let (pair (assoc column changed))
        (cond ((equal (cdr pair) %old-value%)
               ;; If the last changed cell has reverted, clear the row-header indicator
               (unless (setf (changed record) (delete pair changed))
                 (%set-edit-mode grid record nil :redraw-now nil))
               (invalidate-dependents grid record column :redraw-now :header))
              (t ;; else we could operated in :deferred mode
               (invalidate-dependents grid record column :redraw-now t))))
      ;; As the change buffer is empty, clear the row-header indicator
      (%set-edit-mode grid record nil :redraw-now t))
    (when exit
      (setf %old-value% *unbound-value*))))

(defgeneric revert-cell (grid record column)
 ;;; Reverting current cell to the previous value in full invoked by Ctrl+Shift+Z
  ;; CAUTION: Avoid set-cell-value -> cell-on-update.
 (:method (grid record column)
  (undo-cell grid record column))			; beep by default

 (:method :after ((grid ygrid) (record record) column)
  (%after-revert-cell grid record column))

 (:method :around ((grid ygrid) (record record) column)
  (if (record-filled-p grid record) (call-next-method) (beep))) )

(defgeneric record-on-revert (grid record)
 ;;; Undo the effect of editing an maybe of record-on-change. Possibilities:
  ;; - revert all column data (including current column if it was changed several times),
  ;; - set the entire record-data object from a backup copy.
 (:method ((grid ygrid) (record record))
  (dolist (pair (changed record))
    ;(when-let (column (car pair))			; ignore fictive pair?
    (call-data-writer (cdr pair) grid (record-data record) (car pair)))
  ;; Revert all column data maybe except the current column
  #|(loop for (column . value) in (changed record)
          do (call-data-writer value grid (record-data record) column)
             (and (eq current-column column)
                  (setf (slot-value grid '%old-value%) value)))|#; prepare for reverting it
  (remf (record-plist record) :changed)) )

(defgeneric revert-record (grid &key record interactive)
 ;;; Revert current record or current cell
 (:method ((grid ygrid) &key (record (current-record grid)) (interactive t))
  (if (%get-edit-mode grid)
      (when (imp interactive
                 (beep-box #L"Do you really want to revert the current record~@[~%of '~a'~] to saved version?"
                           :format-arguments (list (object-title grid))
                           :owner (capi:top-level-interface grid)
                           :buttons '(:ok :cancel) :default-button :cancel))
        ;; Former (%revert-record grid record)
        ;; NB: Our implementation may invoke some data readers twice:
        ;;	 (1) from inside of revert-cell,
        ;;	 (2) from inside of invalidate-record.
        (unless (record-auto-p grid record)
          (let* ((current-column (%current-column grid))
                 (pair (assoc current-column (changed record))))
            (record-on-revert grid record)
            ;; Prepare for reverting the current cell, then revert it without invalidating
            ;; dependents in the :after method as the changed alist is now empty.
            ;; Q: If (overlay (pinboard grid)), i.e. editing pane is shown,
            ;;    what about popup?!
            (when pair
              (setf (slot-value grid '%old-value%) (cdr pair))
              ;; Prevent undesirable beep - revert current cell only if it's really changed
              (revert-cell grid record current-column)))); does not call cell-on-update!
        (%set-edit-mode grid record nil :redraw-now nil) ; let invalidate-record redraw
        (invalidate-record grid record))

        ;; Record has not changed - try on cell only.  Q: (operate-on-cell-p ... :undo)?
        (revert-cell grid record (%current-column grid)) )) )
     ;(interactive (capi:beep-pane))

#|(defmethod %revert-record ((grid ygrid) (record record))
 ;;; Called only when current record edit mode is on.
  ;; NB: Our implementation may invoke some data readers twice:
  ;;	 (1) from inside of revert-cell,
  ;;	 (2) from inside of invalidate-record.
  (unless (record-auto-p grid record)
    (let* ((current-column (%current-column grid))
           (pair (assoc current-column (changed record))))
      (record-on-revert grid record)
      ;; Prepare for reverting the current cell, then revert it without invalidating
      ;; dependents in the :after method as the changed alist is now empty.
      ;; Q: If (overlay (pinboard grid)), i.e. editing pane is shown -  what about popup?!
      (when pair
        (setf (slot-value grid '%old-value%) (cdr pair))
        ;; Prevent undesirable beep - revert current cell only if it was really changed
        (revert-cell grid record current-column))))	; does not call cell-on-update!
  (%set-edit-mode grid record nil :redraw-now nil)	; let invalidate-record redraw
  (invalidate-record grid record))|#

(defgeneric save-record (grid &key redraw-now &allow-other-keys)
 (:documentation "Save current record if needed.
Values:
 (True NIL)       Success - the first value returned by record-on-update.
 (T True/:ignore) No need to save, or the second comes from save-sell,
                  or an error in record-on-update that has been ignored.
 (NIL any)	  Failed save-cell, the second value comes from it.
 (NIL :abort)     Failed record-on-update, the abort restart has chosen."))

(defmethod save-record ((grid ygrid)
                        &key (redraw-now t)
                             (record (current-record grid))
                             (cell (< (current-col grid) (length (columns grid))))) ;t)
 ;;; Args:  cell  If true, exit current cell first.
  ;;		  If null, called from record-on-exit, which should follow cell-on-exit
  ;;        force? If true, save even if (%get-edit-mode grid) is not set.
  ;; Values: 1) True - success, The one returned by record-on-update if it was non-null;
  ;;         2) T if no need to save;
  ;;            or :abort If the abort restart was chosen after an error 
  ;;               or record-on-update retruned nil.
  (let ((first nil)
        (second t))
    (cond ((record-auto-p grid record)
           (setq first t))
          ((and cell (not (multiple-value-setq (first second)
                              (save-cell grid record (%current-column grid)))))
           ;(and cell (eq (%save-cell grid record (%current-column grid)) :abort))
           (setq second :abort))
          ((%get-edit-mode grid)
           (when (setq first
                       (restart-case (handler-bind ((error #'prompt-for-restart))
                                      (prog1 (record-on-update grid record)
                                        (setq second nil)))
                        (abort () :report (list #L"(Cancel) Revert to editing row ~d."
                                                (1+ (record-row-number record)))
                               ;; If we has called cell-on-exit already, the editing
                               ;; overlay can be detached here.
                               (unless cell
                                 (cell-on-enter grid record (%current-column grid)))
                               (ensure-cell-visible grid)
                               (setq second :abort)
                               nil)
                        (ignore () :report #L"Ignore problems with updating record."
                                (revert-record grid :record record :interactive nil)
                                (setq second :ignore)
                                t)))
             ;(%set-edit-mode grid record nil :redraw-now nil)
             (when redraw-now 
               (invalidate-record grid record))))
           ;(setq second :abort))
           ;(values first second)))
        (t
         (setq first t second t)))
    (values first second)))

(defgeneric set-cell-value (grid row-or-record column-designator new-value
                            &key recursive redraw-now)
 ;;; Called by: As a rule, the call is originated from save-cell.
  ;;            Also by undo-cell and revert-cell.
  ;; CAUTION: Should not change %old-value%!
 (:documentation
 "Args: redraw-now Controls in what way the row should be redrawn:
   nil=don't redraw, :auto=invalidate, t=immediate draw, :header=all incl.row header")

 (:method :around ((grid ygrid) (record record) column new-value
                   &key recursive (redraw-now (not recursive)))
  (declare (ignore column new-value))
  (when (if (and (not recursive) (record-auto-p grid record))
            ;; Invoke current-callback on making addition?
            (%fill-auto-record grid record :redraw-now redraw-now) ; if fails, do not set
            t)
    (call-next-method)))
 #|(:method :before ((grid ygrid) (record record) column new-value
                   &key recursive (redraw-now (not recursive)))
  (declare (ignore column new-value))
  (when (and (not recursive) (record-auto-p grid record))
    ;; Invoke current-callback on making addition?
    (%fill-auto-record grid record :redraw-now redraw-now)))|#

 (:method ((grid ygrid) (record record) (column basic-column) new-value
           &key recursive (redraw-now (not recursive)))
  (when (column-data-writer column)
    ;; Is called with :auto only from
    ;; either	(save-cell ... text-column)
    ;;		 -> write-cell-text, (%write-cell-data-from-pane ...),
    ;;		(%choice-column-on-character ...), 
    ;;		(ywi:commit ywi:choice-popup-interface ...),
    ;; when the edit mode has been guaranteed to be established earlier;
    ;; or	(undo-cell ... popup-column-mixin),
    ;; when it is impossible to determine the new value of edit mode.
    (unless (eq redraw-now :auto)
      (%set-edit-mode grid record t :redraw-now redraw-now))
    ;; We must keep track of the old value at the moment before calling data writer
    (with-slots (%old-value%) grid
      (when (and (eq %old-value% *unbound-value*)
                 (not recursive) (cell-current-p grid record column))
        (multiple-value-bind (%datum% %condition%)	; a la cell-value
            (ignore-errors (call-data-reader grid (record-data record) column))
          (setf %old-value% (if %condition% *unbound-value* %datum%))))
      (let ((pair (unless (assoc column (changed record))
                    (let ((old-value
                           (if (and (not recursive) (cell-current-p grid record column))
                                    ;(eq column (%current-column grid)))
                                    ;(neq %old-value% *unbound-value*))
                               %old-value%
                               (call-data-reader grid (record-data record) column))))
                      (unless (eq old-value *unbound-value*)
                        (cons column old-value))))))
        (call-data-writer new-value grid (record-data record) column)
        ;; Data-writer can even direcly modify another column - it is all the same!
        (when pair
          (push pair (changed record)))))
    (cell-on-update grid record column recursive)
    (when redraw-now
      (invalidate-dependents grid record column :redraw-now redraw-now))))

 (:method ((grid ygrid) (row integer) column-designator new-value
           &key recursive (redraw-now (not recursive)))
  (set-cell-value grid (%row-record grid row)
                  (find-column grid column-designator :errorp t)
                  new-value
                  :recursive recursive :redraw-now redraw-now))

 (:method ((grid ygrid) (record record) column-designator new-value
           &key recursive (redraw-now (not recursive)))
  (set-cell-value grid record (find-column grid column-designator :errorp t)
                  new-value
                  :recursive recursive :redraw-now redraw-now))

 (:method ((grid ygrid) (row integer) (col integer) new-value
           &key recursive (redraw-now (not recursive)))
  (set-cell-value grid (%row-record grid row) (find-column grid col :errorp t)
                  new-value
                  :recursive recursive :redraw-now redraw-now)) )

(defgeneric undo-cell (grid record column)
 ;;; Perform minimum piece of undo, cf.
  ;; - for editing pane, revert only the last gesture,
  ;; - for other kind coincide with revert-cell.
 (:method (grid record column)
  (beep-impossible grid :undo :name (object-title column))))

(defgeneric unload-grid (grid name &key redraw-now columns &allow-other-keys)
 ;;; Remove records and columns if columns is true.
 (:method ((grid ygrid) name &key redraw-now (columns t))
  (declare (ignore name))
  (wipe-grid grid :redraw-now nil)
  (capi:remove-capi-object-property grid 'ywi:loaded)
  (with-slots (pinboard internal-width left-frozen-width scroll-x hscroll-bar) grid
    (when columns
      (setf (columns grid) #()
            (current-displayer pinboard) nil
            internal-width nil
            left-frozen-width nil
            (ncols grid) 0  (ground-ncols grid) 0))
        ;(capi:editor-pane-text text-input)  +null-string+
    (setf scroll-x 0)
    (when hscroll-bar (setf (capi:range-slug-start hscroll-bar) 0)))
  (when redraw-now
    (%invalidate-grid grid))))

(defgeneric wipe-grid (grid &key redraw-now &allow-other-keys)
 ;;; Remove all records, invalidate all columns but do not change them
 (:method ((grid ygrid) &key redraw-now columns)
  (with-slots (pinboard records nrows current-row current-col selected-rows scroll-y) grid
    (when-let (overlay (overlay pinboard))		; hide overlay if any
      (%hide-pane overlay t)
      (setf (overlay pinboard) nil))
    (if-bind (record (record-auto-p grid (1- nrows)))	; move auto-record to the top
      (setf (aref records 0) record			; and reuse it if any
            (current-subscript pinboard) record
            (record-row-number record) 0
            (record-selected record) nil
            (record-%cache% record) nil
            (record-%graphics-cache% record) *unbound-value*
            nrows 1
            (fill-pointer records) 1)
      (setf (current-subscript pinboard) nil
            nrows 0
            (fill-pointer records) 0))
    (allf 0
      current-row current-col scroll-y)
    (setf (current-displayer pinboard) (current-column grid))
    (nilf selected-rows (changed grid))
    (when columns					; invalidate-column is actually
      (loop for column across (columns grid)		; called for the sake of
            do (invalidate-column grid column		; after methods
                                  :graphics nil :items t :redraw-now nil)))
    (when redraw-now
      (%invalidate-grid grid :hide)))))

;;;;;;;;;;;;;;;;;;;;;;;;;; New vs auto vs. filled  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod record-auto-p ((grid ygrid) &optional (row-or-record (current-row grid)))
  (cond ((integerp row-or-record)
         (and (= row-or-record (1- (nrows grid)))
              (>= row-or-record 0)
             ;(addition-maker grid)
              (let ((record (%row-record grid row-or-record)))
                (and (null (record-data record))
                     (getf (record-plist record) :new)
                     record))))
        (row-or-record
         (and (= (record-row-number row-or-record) (1- (nrows grid)))
              ;(addition-maker grid)
              (null (record-data row-or-record))
              (getf (record-plist row-or-record) :new)
              row-or-record))))

(defmethod record-filled-p ((grid ygrid) &optional row-or-record)
 ;;; Opposite to auto-p
  (when-let (record (if (integerp (orf row-or-record (current-row grid)))
                        (if (< row-or-record (nrows grid))
                            (%row-record grid row-or-record))
                        row-or-record))
    (when (record-data record)
      record)))

(defmethod record-new-p ((grid ygrid) &optional row-or-record)
 ;;; Value: True if row-or-record filled new record
  ;;        (and at the very bottom for integer or null row-or-record)
  (when-let (record (if (integerp (orf row-or-record (current-row grid)))
                        (if (= row-or-record (1- (nrows grid)))
                            (%row-record grid row-or-record))
                        row-or-record))
    (and (record-data record) (getf (record-plist record) :new))))

(defun %fill-new-record (grid record)
 ;;; Invoke addition maker
  ;; Used when: Moving to an auto record or on inserting in the middle.
  ;; NB: The addition-maker should return two additional values:
  ;;      2> List of column designators (numbers or names) that has changed,
  ;;         e.g. for which non-null default values are supplied by the maker.
  ;;      3> List of names of slots of the instance returned as the first value;
  ;;         these slot names are collected into the :new property.
  ;;         Or *unbound-value* staying for "explicitly supplied NIL".
  ;;         This is used when the underlied record is created immediatly,
  ;;         inside the addition-maker.
  ;;         (As there is no way to check whether the third value NIL
  ;;         was supplied by the caller explicitly.)
  (when-let (addition-maker (addition-maker grid))
    (multiple-value-bind (data changed new) (let ((*grid* grid))
                                              (funcall addition-maker))
      (when data
        (setf (record-data record) data)
        (when changed
          (setf (changed record) (do ((acc ())
                                      (rest (listify changed) (rest rest)))
                                     ((null rest) acc)
                                   (setq acc (acons (find-column grid (first rest)
                                                                 :errorp t)
                                                    nil acc)))))
        (when new
          (setf (getf (record-plist record) :new) (if (eq new *unbound-value*) nil new)))
        data))))

(defmethod %fill-auto-record ((grid ygrid) record &key (redraw-now t))
 ;;; Fill the specified record by default content and add another auto-record at the end
  ;; Args:  record  Either old auto-record - usually invoked when user jumps into it;
  ;;	            or NIL - only create a new auto-record
  ;; Value: Fresh auto-record
  ;;        or NIL if record is given and %fill-new-record fails.
  ;; NB: The function gets called only if (allow-additions grid).
  (when (addition-maker grid)
    (with-slots (nrows records) grid
      (when record
        (unless (%fill-new-record grid record)
          (return-from %fill-auto-record nil))
        (invalidate-record grid record :redraw-now redraw-now)) ; invoke current-callback?
      (let ((auto-record (funcall (slot-value grid 'record-constructor)
                                  :row-number nrows
                                  :plist (list :new t))))
        (vector-push-extend auto-record records *records-increment*)
        (incf nrows)
        (when redraw-now
          (invalidate-row grid (1- nrows) :redraw-now :all)
          (%reset-bars grid :horizontal nil))
        (update-navigation-toolbar grid)
        auto-record))))

;;;;;;;;;;;;;;;;;;;;;  INVALIDATE data and graphics cache  ;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric invalidate-column (grid column &key data graphics title redraw-now
                                           &allow-other-keys)
 (:method ((grid ygrid) (column column) &key (data t) (graphics t) title (redraw-now t))
  (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (when (or data graphics)
    (let ((col (column-number column)))
      (do-records (record grid)
        (%invalidate-record-cache record col graphics data))))
  (when title
    (setf (column-%cache% column) *unbound-value*))
  (when redraw-now 
    (%invalidate-column grid column :redraw-now :all)))

 (:method ((grid ygrid) column-designator &rest args)
  (apply #'invalidate-column grid (find-column grid column-designator) args)) )

(defun invalidate-record (grid record &key (redraw-now :header))
  (%invalidate-record-cache record nil t t)		; data and graphics
  (invalidate-row grid (record-row-number record) :redraw-now redraw-now))

(defun invalidate-cell (grid record column-designator &key (redraw-now t))
 ;;; Invalidate only the cell itself.
  (let ((col (if (integerp column-designator)
                 column-designator
                 (column-number (find-column grid column-designator)))))
    (%invalidate-record-cache record col t t)		; data and graphics
    (%invalidate-cell grid (record-row-number record) col :redraw-now redraw-now)))

(defun invalidate-dependents (grid record column &key redraw-now)
 ;;; Invalidate the cell itself and all dependents,
  ;; which are the columns calculated in a backward chaining manner.
  (declare (ignore column))				; for now, invalidate entire record
  (%invalidate-record-cache record nil t t)		; data and graphics
  (invalidate-row grid (record-row-number record) :redraw-now redraw-now))

(defun %invalidate-record-cache (record &optional col (graphics nil) (data t))
 ;;; Args: record Can either be displayed or filtered off
  (let ((%cache% (record-%cache% record))
        (%graphics-cache% (record-%graphics-cache% record)))
    (when (and %cache% data)
      (if col
          (setf (svref %cache% col) *unbound-value*)
          (fill %cache% *unbound-value*)))
    (cond ((null graphics))
          ((null %graphics-cache%)
           (setf (record-%graphics-cache% record) *unbound-value*))
          ((neq %graphics-cache% *unbound-value*)	; simple vector
           (if col
               (setf (svref %graphics-cache% col) *unbound-value*)
               (fill %graphics-cache% *unbound-value*))) )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SELECTION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Highlighting and positioning sections and current cell

(defmacro do-selected-records ((record grid &optional (row (gensym "row")) rows)
                               &body body)
 ;;; Args: rows Can contain NILs.
  (lw:rebinding (grid)
    `(dolist (,row ,(or rows `(selected-rows ,grid)))
       (when ,row	 			; in case we failed to remove nulls
         (when-let (,record (row-record ,grid ,row))
           ,@body)))))

(defun selected-records (grid &optional sort)
  (let ((acc ()))
    (do-selected-records (record grid)
      (push record acc))
    (if sort
        (sort acc (ecase sort
                    ((:asc t) #'<)
                    (:desc #'>))
              :key #'record-row-number)
        acc)))

(defgeneric highlight-cell (grid record column &key start end &allow-other-keys)
 (:method (grid record column &key start end)
  (declare (ignore start end)))
 (:method (grid (row integer) column &key (start 0) end)
  (highlight-cell grid (%row-record grid row) column :start start :end end)) )

(defgeneric select-section (grid section &optional cell-p) ;if-selected)
 (:documentation "Highlight section. Arguments:
 CELL-P      If true, move the current cell inside the selection.
 Value: T  If succeeded.")
 ;; cell-p cause for changing current-row and current-callback
 ;; IF-SELECTED T = do nothing and succeed, :deselect = deselect and succeed, NIL = fail
 (:method :around ((grid ygrid) record &optional cell-p)
  (declare (ignore grid cell-p))
  (if (and (enabled grid) record (neq (choice-interaction grid) :no-selection))
      (call-next-method)
      (beep-impossible grid :select-record))))

(defmethod select-section ((grid ygrid) (row integer) &optional cell-p)
  (cond ((< row (nrows grid))
         (select-section grid (%row-record grid row) cell-p))
        (cell-p (beep))))

(defmethod select-section ((grid ygrid) (record record) &optional cell-p)
  ;(and (not (record-selected record))
  (or (record-selected record)
      (and (cell-on-exit grid (current-record grid) (current-column grid))
           (record-on-exit grid (current-record grid))
           (with-slots (selected-rows) grid
             (let ((row (record-row-number record)))
               (when cell-p					; don't ensure visible
                 (move-cell grid row (current-col grid) nil))	; nor focus, i.e. hide
               (setf (record-selected record) t)
               (push row selected-rows)
               (invalidate-row grid row :redraw-now :header)
               t)))))

(defmethod select-section ((grid ygrid) (rows list) &optional cell-p)
 ;;; Dotted pair (car . cdr) corresponds to the diapason [car...cdr]
  (let ((car (car rows))
        (cdr (cdr rows)))
    (cond ((listp cdr)						; proper list
           (loop for row in rows
                 for record = #1=(row-record grid row)
                 always #2=(select-section grid record (and cell-p (= row car)))))
          ((<= car cdr)						; (lo . up)
           (loop for row upfrom car upto cdr for record = #1# always #2#))
          (t							; (up . lo)
           (loop for row downfrom car downto cdr for record = #1# always #2#)))))

(defmethod select-section ((grid ygrid) (section (eql :all)) &optional cell-p)
  (declare (ignore cell-p))
  (and (cell-on-exit grid (current-record grid) (current-column grid))
       (record-on-exit grid (current-record grid))
       (with-slots (selected-rows) grid
         (case (choice-interaction grid)
           (:single-selection
            (when-let (record (current-record grid))
              (cond ((record-selected record)
                     (pushnew (record-row-number record) selected-rows))
                    (t
                     (deselect-section grid :all)
                     (select-section grid (%current-record grid))
                     t))))
           ((:multiple-selection :extended-selection)
            (do-records (record grid :filter :on)		; only shown records
              (if (record-selected record)
                  (pushnew (record-row-number record) selected-rows)
                  (setf (record-selected record) t
                        selected-rows (cons (record-row-number record) selected-rows))))
            t)
           (otherwise
            (beep-impossible grid #L"Select all records"))))
       (progn (%invalidate-grid grid)
         t)))


(defgeneric deselect-section (grid section &optional cell-p)
 (:documentation "Unhighlight section. Arguments:
 CELL-P if true, move the current cell inside the rest of selection."))

(defmethod deselect-section ((grid ygrid) (row integer) &optional cell-p)
  (cond ((< row (nrows grid))
         (deselect-section grid (%row-record grid row) cell-p))
        (cell-p (beep))))

(defmethod deselect-section ((grid ygrid) (record record) &optional cell-p)
  (when (record-selected record)
    (with-slots (selected-rows current-row) grid
      (let ((row (record-row-number record))
            (current-changed nil))
        (deletef selected-rows row :count 1)
        (nilf (record-selected record))
        (when (and cell-p selected-rows (= current-row row))
          (setf current-row (first selected-rows)
                current-changed t))
        (invalidate-row grid row :redraw-now :header)
        (when (and current-changed (setq record (row-record grid current-row)))
          (draw-cell grid record (row-header grid) nil nil)) ))))

(defmethod deselect-section ((grid ygrid) (rows list) &optional cell-p)
 ;;; Dotted pair (car . cdr) corresponds to the diapason [car...cdr]
  (with-slots (selected-rows) grid
    (let ((current-changed nil)
          (current-row (current-row grid))
          (cdr (cdr rows)))
      (if (listp cdr)					; proper list
          (dolist (row rows (setf selected-rows (nset-difference selected-rows rows)))
            #1=(when (and cell-p (= current-row row))
                 (setq current-changed t))
            (when row					; in case we failed to remove nulls
              #2=(when-let (record (row-record grid row))
                   (nilf (record-selected record))
                   (invalidate-row grid row :redraw-now :header))))
          (let ((car (car rows)))
            (if (<= car cdr)
                (loop for row upfrom car upto cdr	; (lo . up)
                      do #1# #2# #3=(deletef selected-rows row :count 1))
                (loop for row downfrom car downto cdr 	; (up . lo)
                      do #1# #2# #3#))))
      (when (and current-changed selected-rows)
        (let ((header (row-header grid))
              record)
          (when (setq record (row-record grid current-row))
            #4=(draw-cell grid record header nil nil))
          (when (setq record (row-record grid (setf current-row (first selected-rows))))
            #4#))))))

(defmethod deselect-section ((grid ygrid) (section (eql :rows))
                             &optional (redraw-now :header))
  (when-let (rows (selected-rows grid))
    (nilf (selected-rows grid))
    (do-selected-records (record grid row rows)
      (nilf (record-selected record))
      (when redraw-now
        (invalidate-row grid row :redraw-now redraw-now)))))

(defmethod deselect-section ((grid ygrid) (section (eql :all)) &optional cell-p)
  (declare (ignore cell-p))
  (loop for column across (columns grid)
        when (column-selected column)
        do (setf (column-selected column) nil)
           (%invalidate-column grid column :redraw-now :header))
  (deselect-section grid :rows :header))

