;;;-*- Mode: Lisp; Encoding: (win32:code-page :id 1251) -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Interface and pinboard drawing and updating

(in-package :ystok.grid)

;;; INTERFACE-ON-CREATE executes code that cannot run inside make-instance but only after
;;;  creating representation. In LW 4.4, create-callback is for this and seems to be
;;;  invoked no matter if the interface is top-level or not.
;;; CAPI BUG: When an interface is embedded into a switchable-layout, the first call to
;;;	      display-callback (or capi::pinboard-pane-display) occurs before
;;;	      the create-callback!

(defmethod interface-on-create ((grid ygrid))
  (with-slots (pinboard column-header row-header #+lispworks4.3 %initialized%) grid
    (unless #-(or lispworks4.3 lispworks4.2) nil
      	    #+(or lispworks4.3 lispworks4.2) %initialized%
      #+(or lispworks4.3 lispworks4.2) (setf %initialized% t)
      (yl:debug-format :grid "~&on-create ~s" grid)
      (component-on-create pinboard)
      ;; Canonicalize row height and column widths
      (setf (slot-value grid 'row-height) (row-height-to-pixels (row-height grid) grid))
      (loop for column across (columns grid)	
            do (setf (slot-value column 'width)
                     (column-width-to-pixels (column-width column) grid column)))
      (component-on-create column-header)
      (component-on-create row-header)
      (when (and (allow-additions grid) (not (record-auto-p grid (1- (nrows grid)))))
        (%fill-auto-record grid nil :redraw-now nil)) )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  NAVIGATION TOOLBAR  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun record-number-callback (grid data
                     &optional (record-number (parse-integer data :junk-allowed t)))
  (if (and record-number (<= 1 record-number (nrows grid)))
      (pinboard-on-move (pinboard grid) nil nil (1- record-number) :record-number)
      (beep (format nil #L"Cannot go to record ~a." data) grid)))

(defmethod update-navigation-toolbar ((grid ygrid))
 ;;; Provide navigation toolbar feedback
 (when-let* ((navigation-toolbar (navigation-toolbar grid))
             (width (capi:simple-pane-visible-width navigation-toolbar)))
   (when (plusp width)
     (let ((buttons (CAPI::TOOLBAR-BUTTON-GROUP-BUTTONS navigation-toolbar))
           (current-row (current-row grid)))
       (setf (capi:simple-pane-enabled (elt buttons (capi:search-for-item
                                                     navigation-toolbar :up)))
              (> current-row 0)
             (capi:simple-pane-enabled (elt buttons (capi:search-for-item
                                                     navigation-toolbar :down)))
              (< current-row (1- (nrows grid)))
             (capi:simple-pane-enabled (elt buttons (capi:search-for-item
                                                     navigation-toolbar :new)))
              (and (addition-maker grid) (allow-additions grid)
                   (not (record-auto-p grid current-row)))
             (capi:text-input-pane-text (record-number-pane navigation-toolbar))
              (if-bind (selected-rows (selected-rows grid))  	; user-friendly number
                (format nil "~d (~d)" (1+ current-row) (length selected-rows))
                (princ-to-string (1+ current-row))) )))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  RESIZE and SCROLL  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
                  
(defmethod print-object ((pinboard record-board) stream)
  (print-unreadable-object (pinboard stream :type t :identity t)
    (when-bind (ygrid (capi:element-interface pinboard))
      (with-slots (nrows ncols) ygrid
        (format stream "~Ax~A" nrows ncols)))))

(defun pinboard-on-resize (pinboard x y width height)
  (declare (ignore x y))
  ;; According to an error reported by Francisco Rivera, 
  ;; pinboard-on-resize gets invoked on startup with width=1 and height=1. 
  ;; This results in error (= 0 nil) later on calling capi:set-geometric-hint
  ;; (or something else?)
  (when (> width 1)
    (let ((grid (capi:element-interface pinboard)))   
      #+(or lispworks4.3 lispworks4.2)
      (interface-on-create interface)		; ensure from: capi:display->on-resize
      ;; Compute scale ratio
      (with-slots (row-height navigation-toolbar bottom-right-corner) grid
        (%reset-bars grid)
        ;; Set page size almost equal to the pane size rounding to a multiple
        ;; of row-height.
        (capi:set-vertical-scroll-parameters pinboard
                            :step-size row-height
                            :page-size (- height (rem height row-height)))

        ;; When vertical scroll bar appears or disappears, arrange horizontal bar
        ;; by means of displaying bottom-right-corner or hiding it (set width to 0).
        (when bottom-right-corner
          (let* ((old-width (capi:simple-pane-visible-width bottom-right-corner))
                 (max (capi:get-vertical-scroll-parameters pinboard :max-range))
                 (new-width
                  (if (and max (> max height)
                           ;;      
                           ;; width -      .
                           (>= (+ width (gp:port-width (row-header grid)))
                               (if navigation-toolbar
                                   (+ (if-bind (records-title (records-title
                                                               navigation-toolbar))
                                        (capi:simple-pane-visible-width records-title)
                                        0)
                                       ;(first (capi:layout-description footer-layout)))
                                      (capi:simple-pane-visible-width
                                       navigation-toolbar)
                                      (capi:simple-pane-visible-width
                                       ;(record-number-pane grid))
                                       (records-count-title navigation-toolbar))
                                      4)
                                   2)))             ; 2 or 4 due to layout x-gap
                      (1- ywi:*vscroll-bar-width*)
                      0)))
            (and old-width (/= old-width new-width) ; old-width is null on display
                 (capi:set-geometric-hint bottom-right-corner
                                          :visible-min-width new-width)) ))))
    (reposition-panes pinboard)))


(defun horizontal-on-scroll (grid bar operation where)
  #-debug (declare (ignore where))
  (with-slots (scroll-x pinboard) grid
    (with-slots (%last-scroll-dimension% %last-scroll-operation%) pinboard
      (let ((new-x (capi:range-slug-start bar)))
        (yl:debug-format :grid "Scroll user ~s~%~4toperation=~s where=~s new-x=~s"
                         bar operation where new-x)
        ;; No need to do anything if the position hasn't changed.
        (when (or (eq %last-scroll-operation% :drag)     ; dropped into place
                  (not (eql scroll-x new-x)))
          (setf %last-scroll-dimension% :horizontal
                %last-scroll-operation% operation
                scroll-x new-x))))
    ;(when (neq operation :drag)
    ;  (capi:display-tooltip pinboard))			; remove tooltip if any
    (pinboard-redraw grid :dragp (eq operation :drag))
    ;(gp:set-graphics-state column-header :transform
    ;                       (gp:make-transform &optional 1 0 0 1 (- scroll-x) 0))
    (column-header-redraw grid)))

(defun pinboard-on-scroll (pinboard dimension operation scroll-value
                           &key interactive (reposition t) &allow-other-keys)
 ;;; Args: interactive True if invoked by scroll bar input.
  ;;       reposition  NIL when auto-scrolled from bulk selection operation.
  ;; CAUTION: Don't scroll programmatically to y which is not a multiple of row-height!
  (multiple-value-bind (new-x new-y)
      ;; We could do the below computation ourselves, but use the CAPI mechanism
      ;; for computing an X/Y position according to the step and page size
      ;; constrained to be within the data range.
      (CAPI::COMPUTE-INTERNAL-SCROLL-POSITION pinboard dimension operation scroll-value)
    (let ((grid (capi:element-interface pinboard))
          (redraw nil))
      (yl:debug-format :grid "Scroll ~:[programmatic~;user~] ~s ~s ~s value=~s Computed (X,Y)=(~s,~s)"
             interactive pinboard dimension operation scroll-value new-x new-y)

      ;; Always check for vertical direction
      (with-slots (scroll-y row-header row-height) grid
        (with-slots (%last-scroll-dimension% %last-scroll-operation%) pinboard
          (when (and (neq dimension :pan) (> scroll-value 0))
            ;; new-y is the position wanted but there may be not enough space
            (setq new-y (min new-y
                         (- (capi:get-vertical-scroll-parameters pinboard :max-range)
                            (capi:get-vertical-scroll-parameters pinboard :slug-size)))))
          ;; Always truncate scroll-y to the nearest row-height
          (setq new-y (- new-y (rem new-y row-height)))
          ;; No need to do anything if the position hasn't changed
          (when (or (eq %last-scroll-operation% :drag)	; dropped into place
                    ;(not (eq %last-scroll-dimension% :vertical))
                    (not (eql scroll-y new-y)))
            (setf %last-scroll-dimension% :vertical
                  %last-scroll-operation% operation
                  scroll-y #+lispworks4.4 (max 0 new-y) #-lispworks4.4 new-y)
            (setq redraw t)
            (capi:set-vertical-scroll-parameters pinboard :slug-position new-y)
            (row-header-on-display row-header))))	; track position in the row header

      ;; Other dimension can also take place, usually programmatically
      (case dimension
        ((:pan :horizontal)
         (when (setq new-x (first-or-self scroll-value))
           ;(capi:display-tooltip pinboard)		; remove tooltip if any
           (with-slots (hscroll-bar scroll-x) grid
             (setf scroll-x new-x
                   redraw t)
             (when hscroll-bar
               (incf (capi:range-slug-end hscroll-bar)
                     (- new-x (capi:range-slug-start hscroll-bar)))
               (setf (capi:range-slug-start hscroll-bar) new-x))
             (column-header-redraw grid)
         ))))
      (when redraw
        (pinboard-redraw grid :dragp (if (eq operation :drag) dimension)
                              :vertical (and interactive (eq dimension :vertical)
                                             scroll-value)
                              :reposition reposition)) )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DRAW  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod CAPI::PINBOARD-PANE-DISPLAY :after ((pinboard record-board) x y width height)
 ;;; Because pinboard-pane-display get called before -on-resize get called from inside
  ;; capi:display or somewhere else when inside a switchable-layout, putting
  ;; (interface-on-create) into the :after method capi:interface-display seems too late.
  #+lispworks4.3 (interface-on-create (capi:element-interface pinboard))
  (if (slot-boundp pinboard '%font-height%)			; already created
      (let ((*background-cleared* t))
        (pinboard-redraw (capi:element-interface pinboard)
                         :expose-x x
                         :expose-y #+lispworks4 y
                                   #-lispworks4 (+ y (scroll-y (capi:element-interface
                                                                pinboard)))
                         :expose-width width
                         :expose-height height
                         :reposition nil))
      (interface-on-create (capi:element-interface pinboard))))

(defun pinboard-redraw (grid &key expose-x expose-y expose-width expose-height
                                  dragp vertical (reposition t))
 ;;; Determine which parts of pinboard visible area (including objects) must be redrawn
  ;; Args: expose-x  Visual coordinate (due to lack of scrolling) or null if from left
  ;;       expose-y  Internal coordinate or null if from top
  ;;       vertical  If true, called from vertical scroll - optimize drawing
  (let ((pinboard (pinboard grid)))
   (with-slots (scroll-x scroll-y) grid
    (when reposition 						; when drag, hide all 
      (reposition-panes pinboard (if dragp :hide reposition)))	; but don't disable
    ;(format t "~&pinboard-redraw: x ~s y ~s width ~s height ~s"
    ;        expose-x expose-y expose-width expose-height)
    (capi:with-atomic-redisplay (pinboard)
      ;; Apply scroll position transform so that the data coordinates at the origin
      ;; of the viewport are drawn at the upper left corner of the visible area.
      (gp:with-graphics-translation (pinboard (- scroll-x) (- scroll-y))
        (cond ((and expose-x expose-width)	; manually translate to internal position x
               (draw-pinboard grid pinboard (+ expose-x scroll-x) (or expose-y scroll-y)
                              expose-width (or expose-height (gp:port-height pinboard))))
              (vertical
               (let ((visible-width (min (gp:port-width pinboard)
                                         (- (internal-width grid) scroll-x)))
                     (visible-height (gp:port-height pinboard)))
                 ;(gp:clear-rectangle pinboard 0 0 visible-width visible-height)
                 (draw-pinboard grid pinboard scroll-x scroll-y
                                visible-width visible-height
                                :vertical vertical)))
              (t                             	; called by on-scroll and others
               (draw-pinboard grid pinboard scroll-x scroll-y
                              (gp:port-width pinboard) (gp:port-height pinboard)))
) ))) ) )

(defun draw-pinboard (grid port left top width height &key vertical)
 ;;; Render pinboard contents from (left, top) to (left+width, top+height)
  ;; Args: left,top Internal coordinates of the area to be drawn.
  ;;       port	    Pinboard itself
  ;;                (or printer-port with appropriate transform applied - not implemented).
  ;;       vertical If >0, do not draw any gridlines (invoked from scrolling downward)
  (declare (fixnum left width)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (with-slots (ncols nrows gridlines row-height scroll-y scroll-x) grid
    ;(format t "~&draw-pinboard: left ~s width ~s top ~s height ~s v ~s"
    ;        left width top height vertical)
      ;#+debug
      ;(when (/= scroll-y (capi:get-vertical-scroll-parameters pinboard :slug-position))
      ;  (warn "!draw-pinboard: scroll-y=~s slug-position=~s" 
      ;        scroll-y (capi:get-vertical-scroll-parameters pinboard :slug-position)))
      ;; Data coordinates at viewport origin.
    (let ((topmost-exposed-row (topmost-exposed-row grid))
          (right (+ left width))
          (bottom (+ top height))
          (internal-width (internal-width grid))
          (internal-height (internal-height grid)))
      (declare (fixnum topmost-exposed-row right internal-width))
      ;(format t "~&draw-pinboard transform: ~S" (gp:graphics-port-transform port))
      ;; Fill the bottom and right area with workspace background color
      (let ((y (* (- nrows topmost-exposed-row) row-height)))
        (when (< y bottom)
          (gp:draw-rectangle port left (+ y scroll-y) right bottom
                             :filled t :foreground *workspace-background*)
          (when-let (workspace-text (workspace-text grid))
            (gp:draw-string port workspace-text
                            scroll-x (+ y scroll-y (%font-ascent% port) 2)
                            :foreground *workspace-foreground*))))
          ;(draw-wrapped-text port '("  1" "  2") port
          ;                scroll-x (+ y scroll-y (%font-ascent% port) 2) 0 0 :left :top
          ;                '(:foreground :gray))
      (when (< internal-width right)
        (gp:draw-rectangle port internal-width top right bottom
                           :filled t :foreground *workspace-background*))
      ;; Draw gridlines if not invoked as a result of vertical scrolling downward
      (unless (and vertical (> vertical 0))
        (let ((left-frozen-ncols (left-frozen-ncols grid)))
          (declare (fixnum left-frozen-ncols))
          (case gridlines
            ((:all :horizontal)
             (when (< left internal-width)
               (loop with right fixnum = (min right internal-width)
                     for row fixnum upfrom topmost-exposed-row below nrows
                     and y upfrom (+ scroll-y row-height -1) below bottom by row-height
                     do (gp:draw-line port left y right y :foreground :lightgray)))))
          (case gridlines
            ((:all :vertical)
             (when (< top internal-height)
               (let ((bottom (min bottom internal-height))
                     (columns (columns grid))
                     (left left)
                     (x -1))
                 (declare (fixnum left x)
                          (type simple-vector columns))
                 (when (plusp left-frozen-ncols)	; draw lines for leftmost area
                   (let ((scroll-x (scroll-x grid))
                         (left-frozen-width (left-frozen-width grid)))
                     (declare (fixnum scroll-x left-frozen-width))
                     (do (;(width (gp:port-width port))
                          (col 0 (1+ col)))			
                         ((>= col left-frozen-ncols))
                       (declare (fixnum width col))
                       (incf x (column-width (svref columns col)))
                       ;(when (<= width x) (return))
                       (let ((x (+ x scroll-x)))
                         (declare (fixnum x))
                         (when (<= right x)
                           (return))
                         (when (<= left x)
                           (gp:draw-line port x top x bottom
                                         :foreground (if (= col (1- left-frozen-ncols))
                                                         :black :lightgray)))))
                     ;(incf left (the fixnum (left-frozen-width grid))))
                     (setq left (the fixnum (max left (+ scroll-x left-frozen-width)))
                           x (1- left-frozen-width))))

                 (do ((col left-frozen-ncols (1+ col)))	; draw lines for main area
                     ((>= col ncols))
                   (declare (fixnum col))
                   (when (<= right (incf x (column-width (svref columns col))))
                     (return))
                   (when (<= left x)
                     (gp:draw-line port x top x bottom :foreground :lightgray))))))
            (otherwise
             (when (plusp left-frozen-ncols)		; draw black vertical for frozen
               (let ((x (1- (left-frozen-width grid))))	; in any case
                 (declare (fixnum x))
                 (when (< x (gp:port-width port))
                   (incf x (the fixnum (scroll-x grid)))
                   (gp:draw-line port x top x bottom :foreground :black))))) )))
           #|(loop with bottom = (min bottom internal-height)
                   for col fixnum upfrom 0 below ncols
                   and column across (columns grid)
                   for x fixnum = (1- (column-width column))
                         then (the fixnum (+ x (column-width column)))
                   while (< x right)
                   when (<= left x)
                     do (gp:draw-line port x top x bottom :foreground :lightgray) )|#
      (decf top (rem top row-height))
      (loop with bottom = (+ bottom row-height)
            for row fixnum upfrom topmost-exposed-row below nrows
            and y upfrom scroll-y below bottom by row-height
            when (<= top y)
            do (draw-record grid (%row-record grid row) port y nil left right))) ))

(defun draw-record (grid record port top
                    &optional gl (left (scroll-x grid)) (right (internal-width grid)))
 ;;; Draw record within graphics translation imposed
  ;; Args: top  Internal vertical position of the record
  ;;       gl   True = draw gridlines on top and right (called from invalidate-row)
  ;;       left	Internal coordinate of the area to draw
  (declare (fixnum left right)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let* ((record-selected (record-selected record))
         (height (cell-visible-height grid))		; takes horizontal gl into account
         (gridlines (gridlines grid))
         (vertical (memq gridlines '(:all :vertical)))
         (columns (columns grid))
         (ncols (ncols grid))
         (left-frozen-ncols (left-frozen-ncols grid))
         (x 0)
         column-width column)
        ;(%inversed-graphics-args% (get-inversed-graphics-args pinboard port)))
    (declare (fixnum height ncols left-frozen-ncols x)
             (type simple-vector columns))
    (when (and gl (memq gridlines '(:all :horizontal)))		; draw h.gridline at bottom
      (gp:draw-line port left (+ top height) right (+ top height)
                    :foreground :lightgray))
    ;; Draw the leftmost frozen columns, the value of left does not matter
    (when (plusp left-frozen-ncols)
      (let ((scroll-x (scroll-x grid))
            (left-frozen-width (left-frozen-width grid)))
        (declare (fixnum scroll-x left-frozen-width))
        (do ((next-x 0)
             (x scroll-x next-x)				; back to visial
             (col 0 (1+ col)))
            ((or (>= col left-frozen-ncols) (< right x)))
        (declare (fixnum next-x x col))
        (setq column (svref columns col)
              column-width (the fixnum (column-width column))
              next-x (the fixnum (+ x (the fixnum column-width))))
        (when (< left next-x)
          (when vertical
            (let ((x (the fixnum (1- next-x))))
              (declare (fixnum x))
              (gp:draw-line port x top x (+ top height)
                            :foreground (if (= col (the fixnum (1- left-frozen-ncols)))
                                            :black
                                            :lightgray))))
          (if (or record-selected (column-selected column))
              (gp:with-inverse-graphics (port)
                (let ((*background-cleared* nil))
                  #1=(draw-cell grid record column port
                                (make-rect x top
                                           (the fixnum (1- column-width)) height))))
              #1#)))
        (setq left (the fixnum (max left (+ scroll-x left-frozen-width)))
              x left-frozen-width)))
        ;(setq left (max left (left-frozen-width grid))) )
    ;; Draw main scrollable columns
    (do ((next-x x)
         (dw (if vertical 1 0))					; take v.gridl.into account
         (col left-frozen-ncols (1+ col)))
        ((or (>= col ncols) (< right x)))
      (declare (fixnum col next-x dw))
      (setq column (svref columns col)
                   column-width (the fixnum (column-width column))
                   next-x (the fixnum (+ x (the fixnum column-width))))
      (when (< left next-x)
        (when (and gl vertical)					; draw v.gridline on right
          (let ((right (the fixnum (1- next-x))))
            (declare (fixnum right))
            (gp:draw-line port right top right (+ top height) :foreground :lightgray)))
        (if (or record-selected (column-selected column))
            (gp:with-inverse-graphics (port)
              (let ((*background-cleared* nil))
                #2=(draw-cell grid record column port
                              (make-rect x top (the fixnum (- column-width dw)) height))))
            #2#))
      (setq x next-x))))

(defun cell-graphics-args (grid record column &aux cache-vector args)
 ;;; TODO: Eliminate duplicated args in accordance with
  ;;       "record first, column last" principle.
  (flet ((calculate-args ()
           (if-bind (graphics-args-function (graphics-args-function grid))
             (append (funcall graphics-args-function grid record column)
                     (column-graphics-args column))
             (or (column-graphics-args column)
                 (if (getf (record-plist record) :deleted)
                     '(:background :lightpink))))))
                     ;#.(color:make-rgb (/ #xFF #1=255.0) #2=(/ #xCC #1#) #2#)
    ;; Null if all columns have not any specific graphics args
    (cond ((null (setq cache-vector (record-%graphics-cache% record)))
           ())
          ((eq cache-vector *unbound-value*)
           (when (setq args (calculate-args))
             (setf cache-vector (make-array (ncols grid)             ; create new cache
                                            :initial-element *unbound-value*)
                   (record-%graphics-cache% record) cache-vector
                   (svref cache-vector (column-number column)) args)))
          ((eq (setq args (svref cache-vector (column-number column))) *unbound-value*)
           (setf args (calculate-args)
                 (svref cache-vector (column-number column)) args)
           (when (and (null args) (every #'null cache-vector))
             ;; Verified that all other are null - drop cache at all.
             (nilf (record-%graphics-cache% record)))
           args)
          (args))))

(defmethod draw-cell :after (grid (record record) (column column) port rect
                             &optional graphics-args)
 ;;; Draw dashed border around the contents of the current cell if no overlay is shown.
  ;; NB: Bold drawing over the cell gridlines looks nicer, but the implementation of
  ;;	 drag-scroll becomes rather taugh.
  ;; BAD: We do not redraw the highligh border immediately after
  ;;      the pinboard loses focus due to switching to another window.
  (when (and (cell-current-p grid record column)
             (not (or (overlay port) (record-selected record) (column-selected column))))
   (rect-bind (left top width height) rect
     (with-cell-mask (mask-left mask-width) grid column left width
       (when (plusp mask-width)  
         (with-translated-mask (port nil :left mask-left :top top
                                         :width mask-width :height height)
           (if (capi:pane-has-focus-p (pinboard grid))
               (gp:draw-rectangle port (the fixnum (1+ left)) (the fixnum (1+ top))
                                  (the fixnum (- width 2)) (the fixnum (- height 2))
                                  :foreground
                                    #-(and lispworks4 win32) :color_highlight
                                    #+(and lispworks4 win32) 'win32:color_highlight
                                  :dashed t :dash '(2 3) :thickness 2)
               (gp:draw-rectangle port (the fixnum (1+ left)) (the fixnum (1+ top))
                                  (the fixnum (- width 2)) (the fixnum (- height 2))
                                  :dashed t :dash '(1 3) :thickness 1))
           ;(gp:draw-rectangle port left top
           ;                   (the fixnum (1- width)) (the fixnum (1- height))
           ;                   :dashed t :dash '(1 3))
           ;(gp:draw-rectangle port (the fixnum (+ left 2)) (the fixnum (+ top 2))
           ;                   (the fixnum (- width 3)) (the fixnum (- height 3))
           ;                   :dashed t :dash '(3 3) :thickness 2)
) )) ) ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  INVALIDATE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun %invalidate-cell (grid row col &key redraw-now) ;border
  (let ((rect (cell-rectangle grid row col))		; internal coordinates
        (column-left-frozen-p (column-left-frozen-p grid col)))
    (with-slots (pinboard scroll-x scroll-y) grid
      (if redraw-now
          ;; Enforce immediate redrawing
          (gp:with-graphics-translation (pinboard (- scroll-x) (- scroll-y))
            (let ((record (%row-record grid row))
                  (column (col-column grid col)))
            #|(when border
                (rect-bind (left top width height) rect
                  (case (gridlines grid)
                    (:all
                     (gp:draw-rectangle pinboard (1- left) (1- top) (1+ width) (1+ height)
                                        :foreground :lightgray))
                    (:vertical)
                    (:horizontal) )))|#
              (when column-left-frozen-p
                (incf (rect-left rect) scroll-x))
              (if (or (record-selected record) (column-selected column))
                  (gp:with-inverse-graphics (pinboard)
                    (let ((*background-cleared* nil))
                      (draw-cell grid record column pinboard rect)))
                                 ;(get-inversed-graphics-args pinboard pinboard)
                  (draw-cell grid record column pinboard rect))))
          ;; Else simply invalidate - let CAPI follow the usual redraw event sequence
          (rect-bind (left top width height) rect
            ;(with-cell-mask (mask-left mask-width) grid column left width
            (gp:invalidate-rectangle pinboard
                                     (if column-left-frozen-p left (- left scroll-x))
                                     (- top scroll-y)
                                     width height)) ))))

(defun invalidate-row (grid row &key redraw-now)  ;except-col
 ;;; Args: redraw-now ::= :header Redraw row header as well
  ;;                      :all    Redraw row header and gridlines as well
  ;; NB: If the record is filtered off, its row-number = -1, so it is not touched.
  (when (<= (topmost-exposed-row grid) row (bottom-exposed-row grid))
    (with-slots (pinboard row-height scroll-x scroll-y) grid
      (let ((top (* row row-height)))
        (if redraw-now                           ; enforce immediate redrawing
            (let ((record (%row-record grid row)))
              (case redraw-now 
                ((:header :all)
                 (draw-cell grid record (row-header grid) nil nil)))
              (gp:with-graphics-translation (pinboard (- scroll-x) (- scroll-y))
                (draw-record grid record pinboard top (eq redraw-now :all))))
            (gp:invalidate-rectangle pinboard 0 (- top scroll-y)
                                     (gp:port-width pinboard) (1- row-height)) )))))

(defun %invalidate-column (grid column &key redraw-now)  ;except-col
 ;;; Args: redraw-now ::= :header Redraw column header as well
  ;;                    || :all    Redraw column header and gridlines as well
  ;(when (<= (topmost-exposed-row grid) row (bottom-exposed-row grid))
  (let ((pinboard (pinboard grid))
        (visual-x (- (column-left grid column) (scroll-x grid)))
                  ;   (if (left-frozen-column-p (column-number column)
                  ;                             (left-frozen-ncols grid))
                  ;       0 (scroll-x grid))))
        (width (1- (column-width column))))
    (cond (redraw-now                           ; enforce immediate redrawing
           (case redraw-now 
             ((:header :all)
              ;(draw-cell grid (column-header grid) column nil nil)))
              (column-header-redraw grid column)))
           (pinboard-redraw grid :expose-x visual-x :expose-width width
                            :vertical t
                            :reposition nil))
               ;(gp:with-graphics-translation (pinboard (- scroll-x) (- scroll-y))
               ;    (draw-record grid record pinboard top (eq redraw-now :all))))
          (t
           (gp:invalidate-rectangle pinboard visual-x 0 width (gp:port-height pinboard)))
) ) )

(defun %invalidate-grid (grid &optional cell-p)
 ;;; Args: cell-p  If true, redisplay current cell and
  ;;		   - either move onto it and optionally set input focus
  ;;                 if cell-p eq :focus,
  ;;		   - or hide the overlay if cell-p eq :hide or for auto-record.
  ;;		   CAUTION: move-cell may invoke cell-on-enter!
  (with-slots (pinboard column-header row-header nrows) grid
    (gp:invalidate-rectangle column-header)
    (cond (cell-p
           (when-let (buffer (buffer pinboard))
             (nilf (editor:buffer-modified buffer)))
           (gp:invalidate-rectangle pinboard)
           (if (or (eq cell-p :hide) (zerop nrows) (record-auto-p grid))
               (when-let (overlay (overlay pinboard))
                 (%hide-pane overlay t)
                 (setf (overlay pinboard) nil))
               (move-cell grid (current-row grid) (current-col grid)
                          t (eq cell-p :focus))))
          (t            				; we do need pane reposition here
           (pinboard-redraw grid)))			; but don't focus on editing panes
    (gp:invalidate-rectangle row-header)
    (%reset-bars grid)
    (update-navigation-toolbar grid)))

(defvar %inside-reset-bars% nil)
(defun %reset-bars (grid &key (horizontal t) (vertical t))
 ;;; Make navigation and scroll bars parameters match grid visible size and content.
  ;; Value: t  If pinboard-redraw was called from inside
  ;; NB: Calling (setf capi:title-pane-text) cause for pinboard-on-resize to be invoked
  ;;	 recursively for other grids displayed by the interface!
  (unless (eq %inside-reset-bars% grid)
    (let ((%inside-reset-bars% grid)
          (redrawn nil))          ; set to t if pinboard redraw was called
      (with-slots (nrows row-height scroll-y row-header pinboard hscroll-bar) grid
        (when-let* ((navigation-toolbar (navigation-toolbar grid))
                    (records-count-title (records-count-title navigation-toolbar)))
          (setf (capi:title-pane-text records-count-title)	; -> pinboard-on-resize!
                (format nil #L"of ~d"
                        (if (record-auto-p grid (1- nrows)) (1- nrows) nrows))))

        (when (and vertical (plusp (row-height grid)))
          (capi:set-vertical-scroll-parameters pinboard		; DI 2011-Jan-25: Was +2
                                               :max-range (* (1+ nrows) row-height))
          ;; The scroll internals set :slug-position to 0 if the bar selfdisappeared,
          ;; so reset it manually
          (when-let (new-scroll-y (capi:get-vertical-scroll-parameters pinboard
                                                                       :slug-position))
            (when (/= scroll-y new-scroll-y)
              ;(warn "!%reset-bars: scroll-y=~s slug-position=~s" scroll-y new-scroll-y)
              (setf scroll-y (- new-scroll-y (rem new-scroll-y row-height)))
              (capi:set-vertical-scroll-parameters pinboard :slug-position scroll-y)
              (pinboard-redraw grid)
              (row-header-on-display row-header)
              (setq redrawn t))))

        (when (and horizontal hscroll-bar)
          (let ((pinboard-width (gp:port-width pinboard))
                (hscroll-bar-width (capi:simple-pane-visible-width hscroll-bar)))
            (when (and pinboard-width hscroll-bar-width
                       (> pinboard-width 1) (> hscroll-bar-width 0))
              (let* ((left-frozen-width (left-frozen-width grid))
                     (end (- (internal-width grid) left-frozen-width))
                     (width (- pinboard-width left-frozen-width)))  ; actually for scroll
                (declare (fixnum left-frozen-width end width))
                (setf (capi:range-end hscroll-bar)
                       end
                      (capi:range-slug-end hscroll-bar)
                       (min end (+ (capi:range-slug-start hscroll-bar) width))
                      ;(capi:scroll-bar-line-size hscroll-bar) *default-column-width*
                      (capi:scroll-bar-page-size hscroll-bar) (max 1 width)))))))
      redrawn)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Positioning and focusing

(defun cell-rectangle (grid &optional (row (current-row grid))
                                      (col (current-col grid)))
 ;;; Args:  row,col  If null, current assumed
  ;; Value: (left top width height)
  ;;		Interior region (exluding border) in internal coordinates.
  ;; NB: The region is not guaranteed to be visible as frozen columns can exist
  ;;	 and/or non-zero scrolling is applied.
  (with-slots (row-height) grid
    (make-rect (loop for c upfrom 0 below col
                     sum (column-width (col-column grid c)))
               (* row row-height)
               (1- (column-width (col-column grid col)))
               (cell-visible-height grid))))

(defun cell-at-position (grid visual-x y)
 ;;; Args: visual-x Visual coordinate inside pinboard
  ;;       y	    Internal (due to automatic scroll) coordinate
  ;; Values: (1) row (2) column
  ;;         (3) dx (4) dy - position relative to the left top corner of the cell;
  ;;	     or NIL if no cell can be found at the (visual-x,y) position
  (declare (fixnum visual-x))
  (multiple-value-bind (row dy) (truncate y (row-height grid))
    (when (and (<= 0 row) (< row (nrows grid)))
      (let ((columns (columns grid))
            (left-frozen-ncols (left-frozen-ncols grid))
            (col 0)
            (left 0)
            column column-width)
        (declare (fixnum left-frozen-ncols col left) (type simple-vector columns))
        ;; Seek the leftmost area: check for visual x
        ;; If some exists at the position, it supersedes the usual column
        (when (plusp left-frozen-ncols)
          (while (< col left-frozen-ncols)
            (when (<= left visual-x #1=(the fixnum (+ left -1
                                            (setq column-width (column-width
                                                   (setq column (svref columns col)))))))
              (return-from cell-at-position (values row column (- visual-x left) dy)))
            (setq left (the fixnum (+ left (the fixnum (the fixnum column-width))))
                  col (1+ col))))
        ;; Seek the main area: check for internal x
        (do ((x (the fixnum (+ visual-x (the fixnum (scroll-x grid)))))
             (ncols (ncols grid)))
            ((>= col ncols))
          (declare (fixnum x ncols col))
          (when (<= left x #1#)
            (return-from cell-at-position (values row column (- x left) dy)))
          (setq left (the fixnum (+ left (the fixnum (the fixnum column-width))))
                col (1+ col))) ))))

(defmethod ywi:displayer-rectangle ((displayer column) &optional record screen)
 ;;; Args: screen If true, return visual coordinates relative to the screen of the grid;
  ;;		  if false, return internal coordinates relative to the pinboard.
  (let* ((grid (capi:element-interface *chooser*))
         (col (column-number displayer))
         (rect (cell-rectangle grid (record-row-number record) col)))
    (declare (fixnum col))
    (with-slots (scroll-x scroll-y) grid
      (when (column-left-frozen-p grid col)
        (incf (rect-left rect) scroll-x))
      (when screen
        (rect-bind (left top width height) rect
          (multiple-value-bind (x y)
              (capi:convert-relative-position *chooser* (capi:element-screen grid)
                                              left top)
            (setq rect (make-rect (- x scroll-x 1)	; align to the left gridline
                                  (- y scroll-y 1)
                                  (+ width 2)		; give external size
                                  (+ height 1)))))))
    rect))

(defun ensure-cell-visible (grid &optional (row (current-row grid))
                                           (col (current-col grid)))
 ;;; Value: True if scroll was invoked, false if has been visible already
  (multiple-value-bind (new-scroll-x new-scroll-y) (find-scroll-to-expose grid row col)
    (cond ((or new-scroll-x new-scroll-y)		; always shows editing pane
           (capi:scroll (pinboard grid) :pan :move (list new-scroll-x new-scroll-y))
           t)
          ((let* ((pinboard (pinboard grid))
                  (overlay (overlay pinboard)))
             (and overlay
                  (eq (current-displayer pinboard) (col-column grid col))
                  (eq (current-subscript pinboard) (%row-record grid row))
                  (not (capi:pane-has-focus-p overlay))
                  (if (capi:simple-pane-enabled overlay) ; if editing pane is active,
                      (capi:set-pane-focus overlay)	 ; just pass focus to it
                      (nilf (overlay pinboard))))	 ; else don't show editing pane
                      ;(rect-bind (x y width height) (cell-rectangle grid row col)
                      ;  (%show-pane overlay x y width height))))
             nil)))))

(defun find-scroll-to-expose (grid row col)
 ;;; Compute scroll-x and scroll-y needed to make cell completely visible and to be passed
  ;; to capi:scroll as a scroll value accompanied by :move as a scroll operation
  (rect-bind (x y width height) (cell-rectangle grid row col)
    (with-slots (pinboard scroll-y row-height) grid
      (let (dx dy)
        (unless (column-left-frozen-p grid col)		;(setq dx nil)
          (let ((scroll-x (scroll-x grid))
                (left-frozen-width (left-frozen-width grid)))
            (declare (fixnum scroll-x left-frozen-width))
            (cond ((< (the fixnum (- x left-frozen-width)) scroll-x)
                   (setq dx (- x left-frozen-width)))
                  ((> (setq dx (- (+ x width) (+ scroll-x (gp:port-width pinboard)))) 0)
                   ;; align left if wider then visible area
                   (setq dx (min (the fixnum (- x left-frozen-width))
                                 (the fixnum (+ scroll-x dx)))))
                  ((setq dx nil)))))
        (cond ((< y scroll-y)
               (setq dy y))
              ((> (setq dy (- (+ y height) (+ scroll-y (gp:port-height pinboard)))) 0)
               ;; Always ceil scroll-y to the next row-height
               (setq dy (- dy (nth-value 1 (ceiling dy row-height))))
               ;; align top if higher then visible area
               (setq dy (min y (+ scroll-y (max dy row-height)))))
              ((setq dy nil)))
        (values dx dy)))))

(defun %reposition-pane (pinboard pane x y width height column right focus)
 ;;; Helper showing the pane at the coordinates and size specified or hide if scrolled out
  ;; Args: pane  Editing pane
  ;;       x,y	 Internal position
  ;;       right Right margin for e.g. image or 3D-button
  ;;       focus If True and the pane is about to be shown, it must also receive
  ;;		 input focus
  (let* ((grid (capi:element-interface pinboard))
         (visual-x (- x (scroll-x grid)))  ; translate into visual coordinates first
         (visual-y (- y (scroll-y grid)))) ; may be nil on calling from resize at startup!
    (yl:debug-format :grid "~&%reposition-pane before: ~s visual (x,y)=(~s,~s)
 ~%~4tfocus=~s pinboard-focus=~s"
                        pane visual-x visual-y
                        (capi:pane-has-focus-p pane) (capi:pane-has-focus-p pinboard))
    ;; If pane overlaps visible area by x-dimension and is inside by y-dimension?
    (if (and (<= (gp:xoverlap visual-x
                              (+ visual-x width)
                              (if (column-left-frozen-p grid column)
                                  0
                                  (left-frozen-width grid))
                              (gp:port-width pinboard)) 0)
             (<= 0 visual-y (+ visual-y height) (gp:port-height pinboard)))
        ;(gp:rectangle-overlap visual-x visual-y (+ visual-x width) (+ visual-y height)
        ;                      0 0 (gp:port-width pinboard) (gp:port-height pinboard))
        (%show-pane pane visual-x visual-y (max (- width right) 0) height focus)
        (%hide-pane pane))
    (yl:debug-format :grid "~&%reposition-pane after: focus=~s pinboard-focus=~s"
                     (capi:pane-has-focus-p pane) (capi:pane-has-focus-p pinboard)) ))

(defmethod reposition-pane (pinboard column pane x y width height
                            &optional (focus (capi:pane-has-focus-p pane)))
  (%reposition-pane pinboard pane x y width height column 0 focus))

(defun %show-pane (pane visual-x visual-y width height focus)
 ;;; NB: capi:set-pane-focus doesn't work when executed in other process
  (capi:with-atomic-redisplay (pane)
    (setf (capi:pinboard-pane-position pane) (values visual-x visual-y)
          (capi:pinboard-pane-size pane) (values width height))
    (unless (capi:simple-pane-enabled pane)
      (setf (capi:simple-pane-enabled pane) t)))
  (when (and focus (not (capi:pane-has-focus-p pane) ))
    (capi:set-pane-focus pane)
    (when (typep pane 'capi:editor-pane)
      (capi:call-editor pane "Refresh Screen") )))		; centering current line
     ;(editor:redisplay)
     ;(editor::redisplay-all)
     ;(editor::redisplay-window pane)
     ;(when (slot-value pane 'capi::cursor-font);signals "bad face" if cursor-font is null
     ;  (capi::draw-the-cursor pane t))

(defun %hide-pane (pane &optional (disable t))
 ;;; Args: disable T  - Hide pane by means of disabling and removing focus;
  ;;			that passes focus to pinboard and prevents further motion events.
  ;;	 	   NIL  Zero its size (called by scrolling) without changing focus.
  ;; NB: As (setf capi:pinboard-pane-size) does not work for option-pane, hide it by
  ;;	 moving it above top edge of the board.
  (cond (disable
         (when (capi:simple-pane-enabled pane)	; (eq disable :force))
           (when (capi:pane-has-focus-p pane)
             (capi:set-pane-focus (capi:element-parent pane)))
           (capi:with-atomic-redisplay (pane)
             (setf (capi:simple-pane-enabled pane) nil
                   (capi:pinboard-pane-size pane) (values 0 0)
                   (capi:pinboard-pane-position pane) (values -10 -50)) )))
        ((/= (nth-value 1 (capi:pinboard-pane-size pane)) 0)
         (setf (capi:pinboard-pane-size pane) (values 0 0))
         (when (eq (type-of pane) 'editing-option-pane)
           (setf (capi:pinboard-pane-position pane) (values -10 -50))) )))

(defun reposition-panes (pinboard &optional operation)
 ;;; Used to show/display the overlay if any
  ;; Args: operation ::= :hide     Hide all editing panes.
  ;;                   | :focus    Hide or show the overlay if any and when show,
  ;;                               switch input focus to it; hide other panes.
  ;;                   | otherwise Hide or show the overlay without focusing
  ;;                               and hide other panes.
  (if (eq operation :hide)
      (dolist (pane (ywi::editing-panes pinboard))
        ;; Don't actually disable nor hide but make size zero
        (%hide-pane pane nil))
      (do* (column pane
                   (grid (capi:element-interface pinboard))
                   (overlay (overlay pinboard))
                   (current (and (allow-edits grid)		; true if bound 
                                 (plusp (nrows grid))		; to current cell
                                 (plusp (ncols grid))
                                 (null (selected-rows grid))
                                 (notany #'column-selected (columns grid))
                                 (eq (current-displayer pinboard)
                                     (setq column (%current-column grid)))
                                 (eq (current-subscript pinboard)
                                     (%current-record grid))))
                   (rest (ywi::editing-panes pinboard) (rest rest)))
           ((null rest))
        (if (and (eq (setq pane (first rest)) overlay) current)
            (rect-bind (x y width height) (cell-rectangle grid)
              (when (column-left-frozen-p grid column)
                (incf x (scroll-x grid)))
              (reposition-pane pinboard column pane x y width height
                               (eq operation :focus)))
            ;; Actually hide and disable - if has focus, switch focus to parent
            (%hide-pane pane)))))

(defmethod move-cell ((grid ygrid) new-row new-col &optional (ensure-visible t) focus)
 ;;;  Move highlight or editing pane to the cell <new-row,new-col>
  ;;  and possibly enter into it but do not switch input focus!
  ;; Args: ensure-visible  If true, try to scroll the pane inside visible area.
  ;;                       If false, hide pane in a gentle manner
  ;;       focus           If true, pass focus to corresponding editing pane
  ;;			   (or to the pinboard itself? has not it already had focus?)
  ;; Value: True = success, nil = either cell-on-exit or record-on-exit signals.
  (with-slots (current-row current-col pinboard row-header) grid
    (let* ((old-row current-row)
           (record (%row-record grid old-row))		; can be NIL for empty grid!
           (old-col current-col)
           (column (col-column grid old-col))
           (invalidate-old-cell t)
           (invalidate-new-cell nil)
           (new-scroll-x nil)
           (new-scroll-y nil)
           overlay)
          ;(invalidate-old-row nil) ; ;(invalidate-new-row nil)
      ;(capi:display-tooltip pinboard)			; remove tooltip if any
      (when (or (/= new-row old-row) (/= new-col old-col))
        (unless (cell-on-exit grid record column)	; should nilf overlay in :around
          (return-from move-cell nil))
        (when column
          (help grid :pointer-documentation-leave (help-key column))))
      (shiftf overlay (overlay pinboard) nil)		; DI 2010-Dec-14: retain overlay
      ;(nilf (current-displayer pinboard) (current-subscript pinboard)
      ;(setf (overlay pinboard) nil)

      ;; When row changes, save (update or insert) the current record in needed
      (cond ((/= new-row old-row)
             (unless (record-on-exit grid record)
               (setf (overlay pinboard) overlay)	; restore overlay if was previously
               (return-from move-cell nil))
             (setf current-row new-row
                   current-col new-col
                   record (%row-record grid new-row)
                   column (col-column grid new-col))
             (record-on-enter grid record))		; enter the new current record
            (t
             (setf current-col new-col
                   column (col-column grid new-col))))

      (setf (current-displayer pinboard) column
            (current-subscript pinboard) record)

      (if (or (ywi::edit-on-enter pinboard)
              (null (ywi:editing-pane-type column)))
          (cell-on-enter grid record column)	; overlay := current ed.pane
          (setq invalidate-new-cell t))
      (let ((help-key (help-key column)))
        (help grid :pointer-documentation-enter
              (if (eq help-key t) (column-title column) help-key)))

      ;; Either reposition pane or let scroll method do it
      (cond (ensure-visible
             (multiple-value-setq (new-scroll-x new-scroll-y)
                 (find-scroll-to-expose grid new-row new-col))
             (cond ((or new-scroll-x new-scroll-y)
                    ;; Scroll is going to redraw everything, so no need to redraw cells
                    (nilf invalidate-old-cell)
                    (capi:scroll pinboard :pan :move (list new-scroll-x new-scroll-y)))
                   ((reposition-panes pinboard (if focus :focus)))))
            (t (reposition-panes pinboard :hide)))		; hide but not disable
      (unless new-scroll-y
        (draw-cell grid (%row-record grid old-row) row-header nil nil)
        (unless (= new-row old-row)
          (draw-cell grid (%row-record grid new-row) row-header nil nil)))
      
      (when invalidate-old-cell (%invalidate-cell grid old-row old-col :redraw-now t))
                                                  ;:border t))
      (when invalidate-new-cell (%invalidate-cell grid new-row new-col :redraw-now t))
      t)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  INPUT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod pinboard-on-move ((pinboard record-board) x y gesture &optional
                             (direction (sys::gesture-spec-data gesture)))
 ;;; Handle keyboard input gestures direcly. Skip disabled or hidden (<= width 1) cells.
  ;; Args: gesture   NIL, gesture-spec, or
  ;;                 row-number    for :direction :record-number
  ;;                 column-number for :direction :column-number
  ;;                 (row-number . column-number) for :direction :pan
  (declare (ignore x y))
  (let ((grid (capi:element-interface pinboard)))
   (with-slots (ncols nrows current-row current-col row-height) grid
    ;(deselect-section grid :all)
    ;(let ((row -1) (col -1))
    ;  (declare (fixnum row col))
    (loop with row = current-row and col = current-col
          and column
          for prev-row = row and prev-col = col
          do
          (case direction
            (:left      (setq col (1- prev-col)))
            (:right     (setq col (1+ prev-col)))
            (:home      (setq col 0
                              direction :right))   ; go opposite direction if disabled
            (:end       (setq col (1- ncols)
                              direction :left))	   ; go opposite direction if disabled
            (:page-left (setq col (max 0 
                          (loop with visible-width = (gp:port-width pinboard)
                                for c from prev-col downto 0
                                for column = (col-column grid c)
                                sum (column-width column) into width
                                do (yl:debug-format :grid "~&columm=~s width=~s"
                                                    column width)
                                while (< width visible-width)
                                finally (return c)))
                              direction :right))   ; go opposite direction if disabled
                        ;(format t "~&col=~s" col))
            (:page-right (setq col (min (1- ncols)
                          (loop with visible-width = (gp:port-width pinboard)
                                for c upfrom prev-col below ncols
                                for column = (col-column grid c)
                                sum (column-width column) into width
                                do (yl:debug-format :grid "~&columm=~s width=~s"
                                                    column width)
                                while (< width visible-width)
                                finally (return c)))
                              direction :left))	   ; go opposite direction if disabled
                         ;(format t "~&col=~s" col))
            (:up        (setq row (1- prev-row)))
            (:down      (setq row (1+ prev-row)))
            (:page-up   (setq row (max (- prev-row
                                          (truncate
                                           (capi:get-vertical-scroll-parameters
                                            pinboard :page-size) row-height))
                                       0)
                              direction :down))	   ; go opposite direction if disabled
            (:page-down (setq row (min (+ prev-row
                                          (truncate
                                           (capi:get-vertical-scroll-parameters
                                            pinboard :page-size) row-height))
                                       (if (record-auto-p grid (1- nrows))
                                           (- nrows 2)	; don't breach the new record
                                           (1- nrows)))
                              direction :up))	   ; go opposite direction if disabled
            (:first     (setq row 0))
            (:last      (setq row (if (record-auto-p grid (1- nrows))
                                      (- (nrows grid) 2)
                                      (1- nrows))))
            (:new       (if (and (addition-maker grid) (allow-additions grid))
                            (progn (unless (record-auto-p grid (1- nrows)) ; something gets
                                     (%fill-auto-record grid nil)) 	   ; wrong with it
                              (setq row (1- nrows)))
                         (beep (list
                                #L"Record additions are not allowed within grid~@[ '~a'~]."
                                (object-title grid))
                               grid)))
            ;(:top       (setq row 0 col 0))
            ;(:bottom    (setq row (1- nrows) col (1- ncols)))
            (:record-number (setq row gesture))
            (:column-number (setq col gesture))
            (:pan (setq row (car gesture)
                        col (cdr gesture)))
            (#.(char-code #\Tab)
             (if (eql (sys::gesture-spec-modifiers gesture) #.sys::gesture-spec-shift-bit)
                 (cond ((> prev-col 0) (setq col (1- prev-col)))
                       ((> prev-row 0) (setq row (1- prev-row)
                                             col (1- ncols))) )
                 (cond ((< (1+ prev-col) ncols) (setq col (1+ prev-col)))
                       ((< (1+ prev-row) nrows) (setq row (1+ prev-row)
                                                      col 0))) ))
          )
          while (and (or (/= row prev-row) (/= col prev-col))	; movement has performed
                     (< -1 row nrows) (< -1 col ncols))		; within valid area
          when (and (column-enabled grid (setq column (col-column grid col)))
                    (not (column-hidden column)))		; not hidden
                       ;(if (member (gridlines grid) '(:all :vertical) :test #'eq) 1 0))
          return (progn
                   (deselect-section grid :all)
                   (if (or (allow-edits grid)
                           (not (capi:capi-object-property grid :select-record)))
                       (move-cell grid row col t t)
                       (select-section grid row t)))
          finally (beep nil pinboard)) )))

(defun record-board-on-gesture (pinboard x y gesture)
 ;;; We do need this to deferentiate Ctrl-a from Ctrl-Shift-A more smootly, 
  ;; Q: Search in the leftmost column a la list-panel?
  (declare (ignore x y))
  (let* ((grid (capi:element-interface pinboard))
         (interface (capi:top-level-interface grid))
         (modifiers (sys::gesture-spec-modifiers gesture)))
   (flet ((try-cell-on-character ()
            (if (allow-edits grid)
                (let ((column (%current-column grid))
                      (character (sys:gesture-spec-to-character gesture)))
                  (if (and column (enabled column)) ;(not(column-read-only grid column)))
                      (if (eql character #\Escape)
                          (cell-on-cancel grid column)
                          (cell-on-character grid column character))
                      (beep)))
                (beep-no-edits grid))))
    (case (sys::gesture-spec-data gesture)
      (#.(char-code #\Return)
       (case modifiers
         (#.sys::gesture-spec-shift-bit
          (when (allow-edits grid)
            (save-record grid)))
         (0
          (case (capi:capi-object-property grid :press-return)
            (:action
             (or (cell-on-action grid)
                 (try-cell-on-character)))	; if the action callback doesn't exist
            (:down
             (pinboard-on-move pinboard 0 0 nil :down))
            (otherwise
             (pinboard-on-move pinboard 0 0 #.(sys::make-gesture-spec (char-code #\Tab)
                                                                      0))) ))))
      (#.(char-code #\Escape)
       (case modifiers
         (0
          (try-cell-on-character))))

      (#.(char-code #\Rubout)		 	; Del key but differs from :delete!
       (case modifiers
         (#.sys::gesture-spec-shift-bit
          (if (operate-on-grid-p interface grid :cut)
              (operate-on-grid interface grid :cut)
              (beep-impossible grid :cut)))
          ;(operate-on-focus pinboard :cut))
         (0
          (if (selected-rows grid)
              ;(when (allow-deletions grid)
              (delete-selected-rows grid)
              (try-cell-on-character)))))
      (#.(char-code #\-)
       (if (eql modifiers sys::gesture-spec-control-bit)
           (delete-record-or-rows grid)
           (try-cell-on-character)))
      (#.(char-code #\+)
       (if (eql modifiers sys::gesture-spec-control-bit)
           ;(when (and (addition-maker grid) (allow-additions grid))
           (pinboard-on-move pinboard 0 0 nil :new)
           (try-cell-on-character)))
      ((#.(char-code #\a) #+Russian #.(char-code #\))
       (if (eql modifiers sys::gesture-spec-control-bit)
           (select-section grid :all)
           (try-cell-on-character)))
      ((#.(char-code #\x) #+Russian #.(char-code #\))
       (if (eql modifiers sys::gesture-spec-control-bit)
           (if (operate-on-grid-p interface grid :cut)
               (operate-on-grid interface grid :cut)
               (beep-impossible grid :cut))
           ;(operate-on-focus pinboard :cut)
           (try-cell-on-character)))
      ((#.(char-code #\c) #+Russian #.(char-code #\))
       (if (eql modifiers sys::gesture-spec-control-bit)
           (if (operate-on-grid-p interface grid :copy)
               (operate-on-grid interface grid :copy)
               (beep-impossible grid :copy))
           ;(operate-on-focus pinboard :copy)
           (try-cell-on-character)))
      ((#.(char-code #\v) #+Russian #.(char-code #\))
       (if (eql modifiers sys::gesture-spec-control-bit)
           (if (operate-on-grid-p interface grid :paste)
               (operate-on-grid interface grid :paste)
               (beep-impossible grid :paste))
           ;(operate-on-focus pinboard :paste)
           (try-cell-on-character)))
      ((#.(char-code #\z) #+Russian #.(char-code #\))
       (if (eql modifiers sys::gesture-spec-control-bit)
           (if (operate-on-grid-p interface grid :undo)
               (operate-on-grid interface grid :undo)
               (beep-impossible grid :undo))
           ;(eql modifiers sys::gesture-spec-shift-bit)		  ; -> revert-cell
           (try-cell-on-character)))
      ((#.(char-code #\f) #+Russian #.(char-code #\))		  ; for display column
       (if (or (eql modifiers sys::gesture-spec-control-bit)	  ; as menu accelerator
               (eql modifiers (logior sys::gesture-spec-shift-bit ; does not work sometimes
                                      sys::gesture-spec-control-bit)))
           (operate-on-grid interface grid 'searcher)
           ;(capi:find-interface 'searcher :grid grid)
           (try-cell-on-character)))

      (:insert
       (case modifiers
         (#.sys::gesture-spec-control-bit
          (if (operate-on-grid-p interface grid :copy)
              (operate-on-grid interface grid :copy)
              (beep-impossible grid :copy)))
          ;(operate-on-focus pinboard :copy))
         (#.sys::gesture-spec-shift-bit
          (if (operate-on-grid-p interface grid :paste)
              (operate-on-grid interface grid :paste)
              (beep-impossible grid :paste)))))
          ;(operate-on-focus pinboard :paste))))
      (:F1
       ;; NB: Last explicitly displayed tooltip is saved in positioned-tooltip instance
       ;;     of the pinboard representation and is displayed later over and over
       ;;     until (capi:display-tooltip pinboard) is called without further arguments.
       (case modifiers
         ((0 #.sys::gesture-spec-shift-bit)		; contextual help "What's this?"
         #|(let* ((column (current-column grid))
                 (help-key (if column (help-key column) nil))
                 (tooltip (ywi:help-key-tooltip help-key)))
            (if tooltip
                (capi:display-tooltip pinboard
                                      :x (- (column-left grid column)
                                            (if (column-left-frozen-p grid column)
                                                0
                                                (scroll-x grid))
                                            -3)
                                      :y (+ y 20)
                                      :text tooltip)
                (capi:display-tooltip pinboard))|#	; remove tooltip
            (when (eql modifiers 0)			; only needed when no help-menu
              (help-callback (capi:top-level-interface grid) pinboard :help
                             (capi:help-key pinboard)))
         )
         (otherwise
          (try-cell-on-character))))
      (:F2
       (case modifiers
         (#.sys::gesture-spec-shift-bit
          (operate-on-focus pinboard :zoom))
         (otherwise
          (try-cell-on-character))))
      (:F5
       (case modifiers
         (0
          (operate-on-grid interface grid :refresh))
         (otherwise
          (try-cell-on-character))))
      ((#.(char-code #\j) #+Russian #.(char-code #\))
       (case modifiers
         (#.sys::gesture-spec-control-bit
          (operate-on-grid interface grid :indent))
         (#.(+ sys::gesture-spec-control-bit #.sys::gesture-spec-shift-bit)
          (operate-on-grid interface grid :unindent))
         (otherwise
          (try-cell-on-character))))

      (otherwise
       (try-cell-on-character))))))

(defun pinboard-on-select (pinboard x y &optional modifier)
 ;;; Args: x - visual coordinate due to lack of hor.scroll - retranslate to internal
  ;;       y - internal translated coordinate for LW4 or visial for LW5 (?!)
  ;;       modifier ::= nil | :control | :shift | :second-press
  (let ((grid (capi:element-interface pinboard)))
    (multiple-value-bind (row column dx dy)
        (cell-at-position grid x #+lispworks4 y
                                 #-lispworks4 (+ y (scroll-y grid)))
        ;(cell-at-position grid (+ x (scroll-x grid)) y)
      (when (and row column)
        (if (enabled column)
            (case modifier
              (:second-press		; always follows the first press
               ;(deselect-section grid :all)
               (when (or (and (eql (current-row grid) row)
                              (eql (current-col grid) (column-number column)))
                         (move-cell grid row (column-number column) t t))
                 ;; Action is possible even without overlay
                 (cell-on-action grid (%row-record grid row) column dx)))
              (otherwise
               ;; Remember the internal time as the :second-press instead of going to
               ;; to pinboard may be captured by an editing-pane as a first press.
               (setf (button1-press-internal-time pinboard) (get-internal-real-time))
               (deselect-section grid :all)
               (if (or (allow-edits grid)
                       (not (capi:capi-object-property grid :select-record)))
                   ;(select-cell grid row col t t)
                   (cell-on-click grid row column dx dy) ; -> move-cell
                   (select-section grid row t))))	 ; validates itself
            (beep))))))


(defun pinboard-on-focus (pinboard status)
 ;;; Actually loses focus on deactivating top level interface
  ;; Pressing Tab inside an editing pane does not pass focus to pinboard
  ;; Exploits activate-callback that otherwise has no value for non-top-level grid
  (when status
    (let* ((grid (capi:element-interface pinboard))
           (activate-callback (capi:interface-activate-callback grid)))
      (when (and activate-callback (not (capi:top-level-interface-p grid)))
        (funcall activate-callback grid t)))))
   ;(grid-on-enter (capi:top-level-interface pinboard) (capi:element-interface pinboard))))

(defmethod ywi:find-pane-to-focus ((grid ygrid) &optional test)
  (ywi:find-pane-to-focus (pinboard grid) test))

;(defun pinboard-post-menu (pinboard x y &optional gesture)
; ;;; STUB: The real thing is in menus.lisp
;  (declare (ignore pinboard x y gesture)))

;;; Handle universal accelerators, which can be imposed on top-level interface

(defmethod operate-on-focus ((pinboard record-board) (operation (eql :page-up)))
  (pinboard-on-move pinboard 0 0 nil operation))
(defmethod operate-on-focus ((pinboard record-board) (operation (eql :page-down)))
  (pinboard-on-move pinboard 0 0 nil operation))

;;; Operate on "pseudo-focused" cells, which have no underlying editing pane.

(defmethod operate-on-focus ((pinboard record-board) operation)
  (let ((grid (capi:element-interface pinboard)))
    (operate-on-cell grid (current-record grid) (current-column grid) operation)))

#|(defmethod operate-on-focus ((pinboard record-board) (operation (eql :undo)))
  (let ((grid (capi:element-interface pinboard)))
    (undo-cell grid (%current-record grid) (%current-column grid))))|#

;;; Help callback of type :tooltip can be automatically invoked for the pinboard
;;; - either directly, when no editing pane yet,
;;; - or from the help-callback for an editing pane.
;;;   ,       tooltip 
;;;  .
;;;   ,    tooltip  editing-pane,  
;;;   help-callback :tooltip  ,  context.

(let ((lw:*handle-warn-on-redefinition* :quiet))
 ;(defmethod CAPI:HELP-KEY ((column column))
 ;  (help-key column))
 (defmethod CAPI:HELP-KEY ((pinboard record-board))
   (let* ((grid (capi:element-interface pinboard))
          (column (current-column grid)))
     (or (and column (help-key column))
         (capi:help-key grid)))))
       ;; Ignore string as it would be used as tooltip and pop-up over overall pinboard
       ;(if (stringp help-key) nil help-key)))))

(defmethod help-callback (interface (pinboard record-board) (type (eql :tooltip))
                          &optional help-key)
 ;;; Ignore as tooltip and pop-up over overall pinboard
  ;; Q: Allow if (capi:capi-object-property grid :select-record) ?
  ;(capi:display-tooltip pinboard)		; remove tooltip
  nil) ;(ywi:help-key-tooltip help-key))
