;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Searcher dialog
;;; NB: It uses its own error handler so there is no need for :top-level-hook

(in-package :ystok.grid)

(capi:define-interface searcher ()
 ((pattern-ring :initform (sys::make-ring 20 'searcher) :allocation :class)
  (last-match :initform :any :allocation :class)   ; last actual search mode, may be :data
  (last-options :initform () :allocation :class) 
  ;; Searcher state
  (grid :accessor grid :initarg :grid :initform nil)
  (column)			; column where to search
  (offset)			; character offset if last search succeeded or nil if not
  (predicate)			; last search predicate
  (print-function))		; last print-function
 (:panes
  (pattern capi:text-input-choice  :mnemonic-title #L"&Find what:"
	:title-args '(:visible-min-width (:character 16))
	:items pattern-ring
	:items-count-function #'sys::ring-length  :items-get-function #'sys::ring-ref
	:test-function #'string=
           ;:max-characters 20 :visible-min-width '(:character 20)
	:selection-callback 'change-search-predicate :callback-type :interface
	:change-callback 'change-search-predicate :change-callback-type :interface)
  (match capi:option-pane  :mnemonic-title #L"&Match:"
	:title-args '(:visible-min-width (:character 16))
	:items #1='((:all   . #L"Whole cell")
                    (:any   . #L"Any part of cell")
                    (:start . #L"Start of cell"))
	:selection (or (position last-match #1# :key #'car) 1)	; translate :data to :all
	:items-count-function #'length  :items-get-function #'elt
	:print-function #'cdr  :data-function 'car
	:selection-callback 'change-search-predicate  :callback-type :interface
	:help-key #L"'Any part' or 'Start of cell' means text comparison.")
  (direction capi:check-button-panel :mnemonic-title #L"Di&rection:"
	:title-args '(:visible-min-width (:character 16)) :title-position :left 
	:items '(#L"Up" #L"Down")
	:mnemonics #-Russian '(0 0) #+Russian '(1 1)
	:help-keys '(nil nil)	 ;'(#L"Move from the current row upward"
                                 ;  #L"Start with the current row then move down")
	:interaction :single-selection
	:selection 1	; down
	:layout-class 'capi:row-layout
	:layout-args `(:gap ,+h-gap+)
	:selection-callback
          (lambda (data)			; change the third option check button text
            (setf (capi:item-text (elt (CAPI::BUTTON-GROUP-BUTTONS options) 2))
                                    (if (equal data #L"Up") #L"From bottom" #L"From top")
                                  offset nil))
	:callback-type :data)
  (options capi:check-button-panel :items	;#L"Search &as formatted"
	  `(,(make-instance 'capi:check-button :mnemonic-text #L"Match &case"
		:selection-callback 'change-search-predicate
		:retract-callback 'change-search-predicate
		:callback-type :interface
		:help-key #L"Make difference between uppercase and lowercase.")
	    ,(make-instance 'capi:check-button :mnemonic-text #L"Search &all columns"
		:enabled nil)		; not implemented
	    ,(make-instance 'capi:check-button :text #L"From top"
		:mnemonic #-Russian 0 #+Russian 2
		:callback-type :interface
		:selection-callback #2=(lambda (searcher)
                                         (setf (slot-value searcher 'offset) nil))
		:retract-callback #2#
		:help-key #L"Start next search from the very top or bottom row."))
	:selection last-options
	:layout-class 'capi:column-layout
	:layout-args `(:internal-border 0 :gap ,+v-gap+))
  (find-next-button  capi:push-button  :mnemonic-text #L"Find &next" ;:mnemonic 0
	:visible-min-width '(:character 11)
	:callback 'find-next  :callback-type :interface
	:default-p t)
  (cancel-button capi:push-button  :text #L"Cancel" ;:mnemonic 0
	:visible-min-width '(:character 11)
	:callback (lambda (searcher) (if (eq searcher (CAPI::CURRENT-DIALOG))
                                         (capi:abort-dialog)
                                         (capi:quit-interface searcher)))
	:callback-type :interface
	:cancel-p t))
 (:layouts
  (left-column capi:column-layout '(match direction)
	:gap +big-v-gap+)
  (bottom-row capi:row-layout '(left-column options)
	:gap +big-h-gap+)
  (main-column capi:column-layout '(pattern bottom-row)
	:gap +v-gap+)
  (right-column capi:column-layout '(find-next-button cancel-button)
	:gap +v-gap+)
  (main-layout capi:row-layout  '(main-column right-column)
	:gap +big-h-gap+ :internal-border +padding+))
 (:default-initargs
  :auto-menus nil
  :layout 'main-layout
  #-lispworks4.3 #-lispworks4.3 :create-callback 'interface-on-create
  :destroy-callback 'interface-on-destroy
  :enable-tooltips *show-tooltips*
  :initial-focus 'pattern		; only when the interface is first desplayed
  #-lispworks4.3 #-lispworks4.3 :window-styles '(:toolbox)
  :help-callback 'help-callback))

(defmethod shared-initialize :after ((self searcher) slot-names &rest initargs
                                     &key &allow-other-keys)
  (declare (ignore slot-names initargs))
  (with-slots (column offset predicate print-function) self
    (when-let (grid (grid self))
      (if (plusp (ncols grid))
          (setf (capi:interface-title self)
                (format nil "~a~@[ '~a'~]"
                        #L"Find in column"
                        (object-title (setf column (current-column grid)))))
          (setf (slot-value self 'column) nil)))
    (nilf offset predicate print-function))) 		;row col

(defun %initialize-searcher (self)
 ;;; Code that must be executed after the interface is just created or reinitialized
  (with-slots (pattern column) self
    (setf (slot-value pattern 'CAPI:HELP-KEY)
          (if (typep column 'check-column)
              (string-append* #L"Enter 'no', 'false', or '0' to find unchecked rows.
Enter another non-empty pattern to find checked rows."
                (when (column-nullable column)
                  '(#\Newline #L"Empty string matches mixed-value/undetermined rows.")))
              nil))))

;;; Q: We maintain a single searcher instance per owner, not process.
;;;    As we pass the :owner initarg, the searcher will close along with the owner.
;;; NB: A searcher interface can be relocated only if it is running in the process given.
;;;     Therefore, none can be reused.  Therefore, we cannot rely on 
;;;     the built-in capi:find-interface method as it always displays
;;;     the new instance in the process of its own, not in the owner's process.
(defmethod capi:interface-match-p ((searcher searcher) &key owner process)
  (cond (owner   (eq (interface-owner searcher) owner))
        (process (eq (CAPI::INTERFACE-PROCESS searcher) process))))

(defmethod capi:interface-reuse-p ((searcher searcher) &key)
  nil)

(defmethod operate-on-grid ((owner capi:interface) (grid ygrid) (op (eql 'searcher)))
  (if-bind (searcher (capi:locate-interface 'searcher :owner owner))
    (progn (reinitialize-instance searcher :grid grid)	; => NIL in delivered application!
      (%initialize-searcher searcher)
      (capi:activate-pane (slot-value searcher 'pattern))) ;(capi:activate-pane searcher)
    (let ((rect (capi:capi-object-property grid 'searcher)))
      (setq searcher (apply #'make-instance 'searcher :grid grid
                            (when rect
                              `(:best-x ,(rect-left rect)
                                :best-y ,(rect-top rect)
                                :best-width ,(rect-width rect)
                                :best-height ,(rect-height rect)))))
      (capi:display searcher :owner owner :process nil))))

#-lispworks4.3
(defmethod interface-on-create ((self searcher))
  (%initialize-searcher self))

#+lispworks4.3
(defmethod capi:interface-display :before ((self searcher))
  (%initialize-searcher self)
  #+win32 (win32:set-window-long (capi:simple-pane-handle self)
                                 ywi::GWL_EXSTYLE ywi::WS_EX_TOOLWINDOW))

(defmethod interface-on-destroy progn ((face searcher))
 ;;; Save searcher window coordinates within the master grid.
  (when-let (grid (grid face))
    (setf (capi:capi-object-property grid 'searcher)
          (multiple-value-list (capi:top-level-interface-geometry face)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun change-search-predicate (searcher &optional value print)
 ;;; When invoked as a callback, drop predicates as well as both print-function and offset
  (with-slots (predicate print-function offset) searcher
    (if (setf predicate value)
        (setf print-function print)
        (nilf print-function offset))))

(defgeneric compile-search-predicate (column text match
                                      &optional searcher case-sensitive-p)
 ;;; Args: text  Pattern string entered by the user.
  ;;	   match Search mode, one of :all, :any, or :start.
  ;; Value: 1> predicate of signature: subject => generalized boolean, i.e.
  ;;		position (for text mode), true (for :data mode), or false;
  ;;	    2> new (specialized) value of match;
  ;;	    3> print function/format string or NIL if text comparison is not applicable.
 (:method (column text match &optional searcher case-sensitive-p)
  (error #L"Cannot search in the column of this type!")))

(defun find-next (searcher)
 ;;; We need handler-case because top-level-hook works only for the master process window.
  ;; If match kind is :data, compare data, otherwise text.
  (declare (optimize (speed 3)))
  (let ((grid (grid searcher))
        (column (slot-value searcher 'column))
        (match (car (capi:choice-selected-item (slot-value searcher 'match)))))
    (when (or (<= (ncols grid) 0) (neq column (%current-column grid)))
      (reinitialize-instance searcher)
      (setq column (slot-value searcher 'column)))
    (unless column
      (capi:beep-pane searcher)
      (return-from find-next))

   (with-slots (pattern direction options find-next-button offset
                last-match last-options) searcher
     (handler-case
      (let* ((text (string-right-trim *whitespace-bag*
                                      (capi:text-input-pane-text pattern)))
             (length (length text))
             (from-scratch (member 2 (capi:choice-selection options)))
             (predicate (slot-value searcher 'predicate))
             (print-function (slot-value searcher 'print-function)))
        (declare (type string text) (fixnum length))
        (cond (predicate
               (setq match last-match))			; recover last mode, may be :data
              ((multiple-value-setq (predicate match print-function)
                   (compile-search-predicate column text match searcher
                                             (member 0 (capi:choice-selection options)))))
              (t (return-from find-next)))
        (change-search-predicate searcher predicate print-function)
        (unless (zerop length)
          (ywi:ring-collection-pushnew text pattern))
        ;; Cycle over rows
        (loop with v-step = (if (eql (capi:choice-selection direction) 0) -1 1)
              and subject
              and condition
              with top = (if (or from-scratch (minusp v-step))
                             0				; up
                             (current-row grid))	; down
              and bottom = (if (or from-scratch (plusp v-step))
                               (- (nrows grid)		; down
                                  (if (record-auto-p grid (1- (nrows grid))) 2 1))
                               (current-row grid))	; up
              for row = (cond (from-scratch
                               (if (plusp v-step) top bottom))
                              ((and (eq match :any) offset) ; (not (zerop offset)))
                               (current-row grid))
                              (t			; if previous search succeeded
                               (+ (current-row grid) (if offset v-step 0))))
                      then (+ row v-step)
              while (<= top row bottom)
              ;when (and (column-enabled grid column) (> (column-width column) 1)) ; not hidden
              do (multiple-value-setq (subject condition)
                     (call-data-reader grid (record-data (%row-record grid row)) column))
              unless condition
              when (setf offset (cond ((eq match :data)
                                       (and (funcall predicate subject) 0))
                                      ((and subject print-function)
				       ;; avoid applying print-function to nil
                                       (funcall predicate
                                                (if (stringp print-function)
                                                    (format nil print-function subject)
                                                    (funcall print-function subject))))))
              do (deselect-section grid :all)
                 (removef (capi:choice-selection options) 2)	; uncheck from-scratch
                 (if (allow-edits grid)
                     (progn
                       (unless (let ((pinboard (pinboard grid)))
                                 (and (eq (current-displayer pinboard) column)
                                      (eq (current-subscript pinboard)
                                          (%row-record grid row))))
                                      ;(overlay pinboard) pane)
                         (move-cell grid row (column-number column)))
                       (if (eq match :all)
                           (highlight-cell grid row column)
                           (highlight-cell grid row column
                                           :start offset :end (+ offset length))))
                     (select-section grid row t))
              and return offset
                     ;(setf (slot-value searcher 'row) row)
              finally (capi:beep-pane searcher)
                      (message-area-format grid #L"No match found in column '~a'."
                                           (column-title column))
                      (setf offset nil))
        (capi:activate-pane find-next-button))
      (error (condition)
             (when (ywi:beep-box "~a" :format-arguments (list condition)
                                 :icon :warn  :owner searcher
                                 :buttons #+debug '(:try :cancel))
               #+debug (invoke-debugger condition))
             (capi:set-pane-focus pattern)))	; end of handler-case

     (setf last-match match			; save the actual mode, not the specified!
           last-options (capi:choice-selection options))) ))

;;; Column class predicate specializations

(defmethod compile-search-predicate ((column text-display-column) text match
                                     &optional searcher case-sensitive-p)
  (declare (type string text))
  (let* ((length (length text))
         (pattern (cond ((member match '(:start :any))) ; text mode -pattern doesn't matter
                        ((zerop length)
                         (if (column-nullable column)
                             (progn (setq match :data) nil)
                             (signal-not-nullable-error (grid searcher) column)))
                            ;(error #L"Column ~s is not nullable - cell value is required."
                            ;      (column-title column))))
                        ((typep column 'text-input-column-mixin)
                         (let ((*grid* (grid searcher))
                               (*record-data* nil)
                               (*column* column))
                           (setq match :data)		; switch to data search mode
                           (funcall (column-data-write-converter column) text))))))
    (declare (fixnum length))
    (when (and (zerop length) (neq match :data))
      (error #L"Empty search string matches any text."))
    (let ((predicate
           (case match
             (:data
              (if pattern
                  (if-bind (test-function (column-test-function column))
                    ;; avoid applying test-function to null subject
                    #1=(lambda (subject) (and subject
                                              (funcall test-function subject pattern)))
                    (lambda (subject) (equal subject pattern)))
                  #'null))
                    #|(case-sensitive-p
                     (if test-function
                         #1=(lambda (subject) (and subject (funcall test-function
                                                                    datum subject)))
                         (lambda (subject) (equal datum subject))))
                    ;; Do not match case - try substitute case-insensitive comparison
                    ((or (eq test-function #'lw:locale-string=)
                         (eq test-function 'lw:locale-string=))
                     (lambda (subject) (lw:locale-string-equal datum subject)))
                    ((or (eq test-function #'string=) (eq test-function 'string=))
                     (lambda (subject) (string-equal datum subject)))
                    ((or (null test-function) 
                         (eq test-function #'equal) (eq test-function 'equal))
                     (lambda (subject) (equalp datum subject)))
                    (t #1#)))|#		; cannot substitute anything reasonable
             (:all
              (if case-sensitive-p
                  (lambda (subject) (declare (type string subject))
                    (if (string= subject text) 0))
                  (lambda (subject) (declare (type string subject))
                    (if (lw:locale-string-equal subject text) 0))))
             (:any
              (if case-sensitive-p
                  (lambda (subject) (declare (type string subject))
                    (search text subject
                            :start2 #3=(if-bind (offset (slot-value searcher 'offset))
                                         (1+ offset) 0)
                            :test #'char=))
                  (lambda (subject) (declare (type string subject))
                    (search text subject :start2 #3# :test #'lw:locale-char-equal))))
             (:start
              (if case-sensitive-p
                  (lambda (subject) (declare (type string subject))
                    (search text subject  :end2 #4=(min length (length subject))
                            :test #'char=))
                  (lambda (subject) (declare (type string subject))                    
                    (search text subject  :end2 #4#  :test #'lw:locale-char-equal)))) )))
      (values predicate match (column-print-function column)) )))

(defmethod compile-search-predicate ((column check-column) text match
                                     &optional searcher case-sensitive-p)
 ;;; Args: text :true  - yes | true | 1 
  ;;            :false - no | false | 0
  ;;		NIL (undetermined for nullable columns only) - ""
  (declare (type string text))
  (let ((nullable (column-nullable column)))
    (values (cond ((zerop (length text))
                   (if nullable
                       #'null			; mixed-value, three-state
                       (error
     #L"For ordinary (two-state) check column the empty string does not match anything.")))
                  ((member text '("yes" "true" "1"
                                  #+Russian "" #+Russian "")
                           :test #'lw:locale-string-equal)
                   (lambda (subject) (and subject
                                          (not (eql subject 0)) (neq subject :false))))
                  ((member text '("no" "false" "0"
                                  #+Russian "" #+Russian "")
                           :test #'lw:locale-string-equal)
                   (if nullable
                       (lambda (subject) #1=(or (eql subject 0) (eq subject :false)))
                       (lambda (subject) (or (null subject) #1#)))))
            :data
            nil)))
