;;;-*- Mode: Lisp; Encoding: (win32:code-page :id 1251) -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; General text-column functionality and CAPI Editor-based text input

(in-package :ystok.grid)

(defclass text-display-column (column)
 (;(header-align :initform nil)
  (align :accessor column-align :initarg :align :initform :left)
  ;; cell-value -> cell-text: func or format-spec; to be applied to a non-nil argument.
  (print-function :accessor column-print-function :initarg :print-function
                  :initform 'princ-to-string)
  ;; Test predicate returns false when the altered value conceptually differs from the old.
  ;; If null:
  ;; - during search or filtering, data is compared via equal,
  ;; - when finding in pop-up collection by means of ywi:find-data-in-collection,
  ;;   data is compared via equal
  ;;   (data within the collection itself is compared via (getf pane-args :test-function),
  ;; - for changeble column subclasses, any altering is considerd as a substantial change.
  ;; The function is always called on data arguments both non-null.
  ;; Signature: (subject pattern), where
  ;;	subject  is a new value or collection item data or column data of records,
  ;;    pattern  is the old value or the value we are searching for or filtering on.
  (test-function :accessor column-test-function
                 :initarg :test-function :initform nil)))

(defmethod column-completion-function (column)
 ;;; Args: (column column) would preceede (column text-input-column-mixin)!
 nil)

(defmethod initialize-instance :after ((column text-display-column) &key)
  ;; Ensure default-initarg (:enabled :read-only) for direct instances
  ;; of text-display-column (but not of a subclass!)
  (when (and (enabled column) (eq (type-of column) 'text-display-column))
    (setf (slot-value column 'enabled) :read-only)))
  ;(unless (column-header-align column)		; DI 2010-Jun-08: Use default alignment
  ;  (setf (column-header-align column) (column-align column))))

;;; HELPERS

(defun %display-column-text (datum print-function)
  (cond ((null datum) nil)          ;(eq datum *unbound-value*)?
        ((typep datum 'condition)
         (format nil "#<~a>" (class-name (class-of datum))))
        ((stringp print-function) (format nil print-function datum))
        ((funcall print-function datum))))

(defun wrap-column-text (grid-component datum column port width height
                         &optional font
                                   (print-function (column-print-function column)))
 ;;; Args: width and height : Internal size with taking padding into accout
  (declare (fixnum width height)
           #-debug (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let ((font-height (the fixnum (1- (if font		; condense vertically a bit
                                         (gp:get-font-height port font)
                                         (%font-height% grid-component)))))
        (font-ascent (if font
                         (gp:get-font-ascent port font)
                         (%font-ascent% grid-component))))
    (declare (fixnum font-height font-ascent))
    (let ((*print-right-margin* (values (floor width         ; actually wrap
                                               (gp:get-font-average-width port font))))
          (*print-lines* (values (ceiling (the fixnum (+ font-height font-ascent))
                                          height))))
      ;; Anyhow, apply capi:wrap-text due to possible linefeeds in the text.
      ;; Calculate width in characters needed to call wrap-text accordingly.
      (let ((text (%display-column-text datum print-function)))
               #|(cond ((null datum) nil)          ;(eq datum *unbound-value*)?
                        ((typep datum 'condition)
                         (format nil "#<~a>" (class-name (class-of datum))))
                        ((stringp print-function) (format nil print-function datum))
                        ((funcall print-function datum)))))|#
        (setq text (capi:wrap-text text
                     (cond ((< height (the fixnum (+ font-height font-ascent)))
                            array-dimension-limit)
                            ; (gp:port-string-width pinboard text) 
                           ((<= (nth-value 2 (gp:get-string-extent port text font))
                                width)
                            array-dimension-limit)         ; actually wrap
                           (t (max (values (truncate width (gp:get-font-average-width
                                                            port font)))
                                   1)))))
        ;; Convert singleton list into the element itself
        (if (rest text) text (first text)) ))))


(defmethod compute-natural-size ((grid ygrid) (column text-display-column))
  #-debug (declare (optimize (speed 3) (safety 0) (debug 0)))
  (let* ((pinboard (pinboard grid))
         (print-function (column-print-function column))
         (font (getf (column-graphics-args column) :font))
         (value (reduce-grid #'max grid
                 :key (lambda (data)
                        (multiple-value-bind (%datum% %condition%)
                            (ignore-errors (call-data-reader grid data column))
                          (when (and %datum% (not %condition%))
                            (nth-value 2 (gp:get-string-extent
                                          pinboard
                                          (if (stringp print-function)
                                              (format nil print-function %datum%)
                                              (funcall print-function %datum%))
                                          font))))))))
    (if value (+ value 4) nil)))				; add padding
      
;;; VIEW
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod draw-cell (grid record (column text-display-column) port rect
                      &optional (graphics-args (cell-graphics-args grid record column)))
 ;;;  Draw within the translation established
  ;; Args: rect Internal coordinates. 
  ;; NB: Since the below are not controlled by an outer gp:with-graphics-translation,
  ;;     don't forget to translate left-top position for them:
  ;;	- graphic mask
  ;;	- gp:clear-rectangle call
  ;; NB2: Unfortunatly, we must cons a fresh mask list for every gp:draw-string call
  ;;      due to this method is probably called from inside capi:with-atomic-redisplay
  #-debug (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (with-slots (pinboard) grid
   (rect-bind (left top width height) rect
    (let ((background (getf graphics-args :background))
          (text (record-cache-value (grid record column)
                  (wrap-column-text pinboard (or %condition% %datum%) column
                                    port (the fixnum (- width 2)) height
                                    (getf graphics-args :font)))))
      (with-cell-mask (mask-left mask-width) grid column left width
        (cond (background
               (gp:draw-rectangles port (ywi:make-rect mask-left top mask-width height)
                                   :filled t :foreground background))
              ((not *background-cleared*)
               (apply #'gp:clear-rectangle port	;(transform-mask port rect)
                      (translate-mask port rect
                                      :left mask-left :top top :width mask-width))))
        ;; Provide paddings: 1-pixel left and 2-pixel right (DI 2013-Feb-24: right was 1)
        (with-translated-mask (port rect :left (the fixnum (1+ mask-left)) :top top
                                         :width (the fixnum (- mask-width 3)))
          (draw-wrapped-text pinboard text port 
                             (the fixnum (1+ left)) top width height
                             (column-align column) :top
                             graphics-args)) )))))

(defmethod highlight-cell (grid (record record) (column text-display-column)
                           &key start end)
  (declare (ignore start end))
  (let* ((col (column-number column))
         (rect (cell-rectangle grid (record-row-number record) col)))
    (with-slots (pinboard scroll-x scroll-y) grid
      (when (column-left-frozen-p grid col)
        (incf (rect-left rect) scroll-x))
      (gp:with-graphics-translation (pinboard (- scroll-x) (- scroll-y))
        (gp:with-inverse-graphics (pinboard)
          (let ((*background-cleared* nil))
            (draw-cell grid record column pinboard rect)))))))

(defmethod cell-clip-formats ((column text-display-column))
  '(:value :string))

(defmethod cell-clip ((grid ygrid) (record record) (column text-display-column))
 ;;; Used by: Edit/select
  ;; Sometimes it is not reliable to use print-function due to possible loose
  ;; of digits after the decimal point.
  (multiple-value-bind (datum condition)
      (ignore-errors (call-data-reader grid (record-data record) column))
    (cond (condition	nil)
          ((null datum)	nil)
          (t
           (let ((print-function (column-print-function column)))
             (values (if (stringp print-function)
                         (format nil print-function datum)
                         (funcall print-function datum))
                     (list :value datum)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;  TEXT-INPUT-COLUMN-MIXIN  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Abstract and editor-based text input columns

(defclass text-input-column-mixin ()
 (;; cell-value -> text-input-buffer (string); not needed if column is disabled
  (data-read-converter :accessor column-data-read-converter
                       :initarg :data-read-converter
                       :initform 'text-read-converter)
  ;; data-write-converter: text-input-buffer -> cell-value, e.g. read-from-string
  ;; If null, text buffer is not writeable
  ;; Buffer text is trimmed from the right before being passed to data-write-converter.
  ;; Serves as on exit -      ,   ..
  ;; Is not invoked for a nullable column if the buffer text is empty after trimming.
  (data-write-converter :accessor column-data-write-converter
                        :initarg :data-write-converter
                        :initform 'identity) ;'text-write-converter
  ;; Signature: (grid record column text)
  (completion-function :reader column-completion-function
                       :initarg :completion-function :initform nil)
) )

#|(defmethod shared-initialize :after ((column text-input-column-mixin) slot-names
                                     &key print-function data-read-converter)
 ;;; Shall we default read-converter to print-function?
  (when (and print-function (not (stringp print-function)) (not data-read-converter))
    (setf (column-data-read-converter column) print-function)))|#

(defun read-cell-text (grid record column)
  (with-slots (%old-value%) grid
    (multiple-value-bind (datum condition)
        (ignore-errors (call-data-reader grid (record-data record) column))
      (if condition
          (prog1 (princ-to-string condition)
            (setf %old-value% *unbound-value*))
          (prog1 (funcall (column-data-read-converter column) datum)
            (setf %old-value% datum))))))

(defun write-cell-text (grid record column text &key (redraw-now :auto))
 ;;; Save current text buffer as datum to cell at the current position
  (let* ((*record-data* (record-data record))		; can be referred in test function
         (datum (if (zerop (length text))
                    (if (column-nullable column)
                        nil
                        (signal-not-nullable-error grid column))
                       ;(error #L"Column ~s is not nullable - cell value is required."
                       ; (column-title column)))
                    (let ((*grid* grid)
                          (*column* column))
                      (funcall (column-data-write-converter column) text)))))
    (with-slots (%old-value%) grid	;text-input buffer
      (when (or (eq %old-value% *unbound-value*)
                (and (neq datum %old-value%)
                     (or (xor datum %old-value%)
                         (null (column-test-function column))
                         (not (funcall (column-test-function column) datum %old-value%)))))
        (set-cell-value grid record column datum :redraw-now redraw-now)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  TEXT-COLUMN  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass text-column (text-display-column text-input-column-mixin) ())

(defclass editing-text-pane (capi:editor-pane) ()
 (:default-initargs
  :visible-min-width nil  :visible-max-width nil
  :visible-min-height nil :visible-max-height nil
  ;:visible-min-width 0 :visible-min-height 0
  :x 0 :y 0
  :width 0 :height 0
  :enabled nil
  :plist '(:title :parent :realm :parent)			; for operate-on-focus
  :focus-callback 'pinboard-on-focus
  ;:pane-can-scroll nil
  :vertical-scroll nil
  :visible-border nil
  :help-key :parent))

(defmethod ywi:editing-pane-type ((displayer text-column)) 'editing-text-pane)

(defmethod ywi:editing-pane-font ((grid ygrid) type column)
  (editing-pane-font (pinboard grid)
                     (if (eq type 'editing-text-pane) 'capi:editor-pane type)
                     column))

(defmethod ywi:make-editing-pane ((type (eql 'editing-text-pane)) (pinboard record-board)
                                  (displayer text-column))
 ;;; CAPI documentation says:
  ;;  If BUFFER-NAME is not passed, then the editor-pane uses some existing buffer.
  ;; Q: Shall we use :buffer-name instead of the explicitly created buffer?
  (let* ((buffer (editor:make-buffer (format nil "grid-~a" (get-internal-run-time))
                                                ;(get-universal-time)
                                     :flag t))	; exclude from the Editor buffer list
         (pane (make-instance type
                 :buffer buffer   ; to be able to specialize on grid instead of pinboard
                 :font (editing-pane-font (capi:element-interface pinboard)
                                          type displayer))))
                 #|:visible-min-width nil  :visible-max-width nil
                 :visible-min-height nil :visible-max-height nil
                 ;:visible-min-width 0 :visible-min-height 0
                 :x 0 :y 0
                 :width 0 :height 0
                 :enabled nil
                 :plist '(:title :parent)
                 :focus-callback 'pinboard-on-focus
                 ;:pane-can-scroll nil
                 :vertical-scroll nil :visible-border nil)))|#
    (initialize-editor-buffer pane (setf (slot-value pinboard 'buffer) buffer))
          ;(slot-value grid 'text-input) text-input)))
    pane))

;;; CONTROLLER
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod cell-on-character ((grid ygrid) (column text-column) character)
  (if-bind (record (current-record grid))
    (progn 
      (deselect-section grid :all)
      (move-cell grid (record-row-number record) (column-number column) t t)
      (let* ((pinboard (pinboard grid))
             (overlay (overlay pinboard)))
        (cond ((null overlay)
               (cell-on-enter grid record column)
               (reposition-panes pinboard))
              ((eql character #\Return))	; don't pass the Return hit to the editor
               ;(#\Control-\Space
              ; (complete-cell grid (%current-record grid) column))
              ((column-read-only grid column)
               (beep))
              (t
               (%set-edit-mode grid record t)
               (capi:call-editor overlay character)))))
    (beep)))

(defmethod cell-on-click ((grid ygrid) row (column text-column) x y)
 ;;; Pass the click gesture to the editor pane
  ;; NB: Releasing mouse capture helps with highlighting but does not help with
  ;;     detecting double-click automatically.
  (let* (;(text-input (text-input grid))
         (pinboard (pinboard grid))
         (overlay (overlay pinboard)))
    (and ;(eq (capi:capi-object-property text-input 'record) (%row-record grid row))
         ;(eq (capi:capi-object-property text-input 'column) column)
         overlay
         (eq (current-displayer pinboard) column)
         (eq (current-subscript pinboard) (%row-record grid row))
         ;(eq (overlay pinboard) text-input)
         (progn 
          ;(setf (button1-press-internal-time (pinboard grid)) (get-internal-real-time))
           #+win32 (win32:release-capture)
           ;(capi::mouse-editor-call text-input x y #'editor-press)))))
           (capi::mouse-editor-call overlay x y #'capi::editor-start-region)))
    t))

(defmethod cell-on-enter ((grid ygrid) (record record) (column text-column))
 ;;; Load text-input buffer by the datum the from specified record and column
  (when (allow-edits grid)
    (let* ((pinboard (pinboard grid))
           (pane (ywi:ensure-editing-pane pinboard column))
           (buffer (capi:editor-pane-buffer pane)))
      (setf (capi:editor-pane-text pane) (read-cell-text grid record column)
            (editor:buffer-modified buffer) nil
            (editor:buffer-writable buffer) (and (not ;(or condition
                                                        (column-read-only grid column))
                                                 (column-data-write-converter column))
            (overlay pinboard) pane)
      (capi:call-editor pane "Clear Undo"))))

#|(defmethod cell-on-exit (grid record (column text-column))
  (when (neq (save-cell grid record column) :abort)
    ;(nilf (capi:capi-object-property (text-input grid) 'record)
    ;      (capi:capi-object-property (text-input grid) 'column))
    t))|#

(defmethod complete-cell ((grid ygrid) (record record) (column text-column))
  (if-bind (completion-function (column-completion-function column))
    (let* ((overlay (overlay (pinboard grid)))
           (string (if overlay
                       (capi:editor-pane-text overlay)
                       (read-cell-text grid record column))))
      (flet ((complete-text (value &optional text save enter)
               ;; Args: save  True = Call save-cell afterwards
               ;;       enter True=Simulate cell-on-enter and don't try to save its value
               (when overlay
                 (let ((buffer (capi:editor-pane-buffer overlay)))
                   (setf (capi:editor-pane-text overlay)
                         (or text (funcall (column-data-read-converter column) value)))
                   (when (editor:buffer-writable buffer)
                     (setf (editor:buffer-modified buffer) (not enter)) ;t) DI 2009-May-11
                     (capi:call-editor overlay "End Of Buffer Cancelling Selection"))))
               (when (and (or save (not overlay)) (not enter))
                 (save-cell grid record column :value value :text text))))
        (multiple-value-bind (value text) (funcall completion-function grid record column
                                                   string)
          ;; Q: If the comletion function pops up a non-modal choice interface, 
          ;;    shall we foresee editor buffer update, e.g. by means of cell-on-enter ??
          (cond ((stringp text)
                 (complete-text value text))
                ((stringp value)
                 (complete-text value value (eq text :save) (eq text :enter)))
                ((consp value)
                 (if (rest value)		; more than one-element list
                     (when-let (completion (capi:prompt-with-list value
                                  #L"Choose one of the cell text completions:"
                                  :pane-args
                                  '(:visible-min-height (min (:character 20)
                                                             (+ (:character 1)
                                                                :text-height))
                                    :visible-max-height t)
                                    ;:font #.(gp:make-font-description :family "courier")
                                  :popup-args `(:title ,(interleave-seq 'string "  "
                                                          (list (object-title grid)
                                                                (column-title column)))
                                                ,@+popup-args+)))
                       (complete-text completion nil (eq text :save)))
                     (complete-text (first value) nil (eq text :save)))) ; DI 2008-Sep-29
                (value
                 (complete-text value nil (eq text :save)))
                ((neq text :cancel)			; has already canceled by user
                 (beep)
                 (message-area-format grid #L"No completion for ~s." string)
                 nil)
    ) ) ) )
    (call-next-method)))	; beeps as completion function is undefined

(defun editor-press (pane x y)
 ;;; Try to detect double click and invoke cell-action-callback first. 
  ;; If this callback is not specified or not a double click, proceed in standard way.
  ;; Called by: capi::mouse-editor-call
  ;; NB: Marking form looks not very accurate due to font difference
  (if (<= (- (get-internal-real-time)
             (button1-press-internal-time (capi:element-parent pane)))
          *double-click-duration*)
      ;; Yes, the double click it is!
      (or (cell-on-action (capi:element-interface pane) nil nil x)
          #|(let ((grid (capi:element-interface pane)))
                  (column (%current-column grid))
                  (cell-action-callback (slot-value column 'cell-action-callback)))
               (if cell-action-callback
                   (funcall cell-action-callback grid (%current-record grid) column
                            (if x
                                (let ((size (compute-natural-size grid column)))
                                  (if (iplusp size) (/ x size) nil))
                                nil))|#
          (progn (capi::editor-start-region pane x y)		; default handlers
            (capi::editor-start-mark-form pane x y)))
      (capi::editor-start-region pane x y)))			; default handler

(defun editor-second-press (pane x y)
 ;;; Try cell-action-callback first. If not specified, proceed in standard way.
  ;; Called by: capi::mouse-editor-call
  ;; TODO: Compute natural size from the buffer conlents instead of the cell value.
  (or (cell-on-action (capi:element-interface pane) nil nil x)
      (capi::editor-start-mark-form pane x y)))			; default handler

(defmethod %write-cell-data-from-pane (grid record (column text-column)
                                       (pane capi:editor-pane))
  (let ((buffer (capi:editor-pane-buffer pane)))
    (when (editor:variable-value-if-bound 'editor:highlight-active-region :buffer buffer)
      (editor::set-highlight-buffer-region nil buffer))) ;=(capi:call-editor "hl off")
    (write-cell-text grid record column
                     (string-right-trim *whitespace-bag* (capi:editor-pane-text pane))))
         
(defmethod save-cell ((grid ygrid) (record record) (column text-column)
                      &key (value *unbound-value*) text (redraw-now :auto))
 ;;; Args: value  Alternative value to be set instead of editor buffer
  ;;              (when called by complete-cell)
  ;;       text   NIL or string - alternative value when called by set-cell-clip
  ;;		  or complete-cell
  (let* ((pinboard (pinboard grid))
         (pane (overlay pinboard)))
    (if (or (neq value *unbound-value*)
            text
            (and pane
                 (changed pane)
                 (eq (current-displayer pinboard) column)
                 (eq (current-subscript pinboard) record)))
        (progn (%set-edit-mode grid record t :redraw-now redraw-now)
          (restart-case
              (handler-bind ((error #'ywi:prompt-for-restart))
                (cond ((neq value *unbound-value*)
                       (with-slots (%old-value%) grid
                         (when (or (eq %old-value% *unbound-value*)
                                   (and (neq value %old-value%)
                                        (or (xor value %old-value%)
                                            (null (column-test-function column))
                                            (not (funcall (column-test-function column)
                                                          value %old-value%)))))
                           (set-cell-value grid record column value
                                           :redraw-now redraw-now)
                           (when pane
                             (setf (changed pane) nil
                                   (capi:editor-pane-text pane)
                                    (or text
                                        (funcall (column-data-read-converter column)
                                                 value))))
                           (values t nil))))
                      (text
                       (write-cell-text grid record column	; -> set-cell-value
                                  (setq text (string-right-trim *whitespace-bag* text))
                                  :redraw-now redraw-now)
                       (when pane
                         (setf (changed pane) nil
                               (capi:editor-pane-text pane) text))
                       (values t nil))
                      (t
                       (%write-cell-data-from-pane grid record column pane)
                       (setf (changed pane) nil)
                       (values t nil))))
            (abort ()
             :report (list
                      #L"(Cancel) Revert to editing the previous cell at row ~d column ~s."
                      (1+ (record-row-number record))
                      (column-title column))
             (ensure-cell-visible grid (record-row-number record) (column-number column))
             (values nil :abort))
            (ignore ()
             :report #L"Ignore last modifications made to the cell."
             (revert-cell grid record column)			; DI 2011-Apr-06
             (values t :ignore)) ))
        ;; Else nothing changed
        (values t t))))

(defmethod highlight-cell (grid (record record) (column text-column) &key (start 0) end)
  (declare (ignore record))
  (let* ((pinboard (pinboard grid))
         (overlay (overlay pinboard)))
    (when overlay
      (let ((buffer (capi:editor-pane-buffer overlay))) ;  (buffer pinboard)))
        (editor:beginning-of-buffer-command nil)
        (cond (end 
               (let ((point (editor:buffer-point buffer)))
                 (editor:character-offset point start)
                 (editor:set-current-mark point)
                 (editor::set-highlight-buffer-region t buffer)
                 (editor:character-offset point (- end start))))
              (t ;(eq match :all)
               (editor:set-current-mark (editor:buffer-point buffer))
               (editor::set-highlight-buffer-region t buffer)
               (editor:end-of-buffer-command nil)) )))))

;;; Edit/select

(defmethod cell-clip ((grid ygrid) (record record) (column text-column))
 ;;; This could be based on text-input-column-mixin instead of text-column
  ;; unless text-display-column preceeded text-input-column-mixin in many subclasses!
  ;; So we must add a similar method on text-choice-pane-column if neeeded.
  ;; A la read-cell-text, but does not return error message.
  (multiple-value-bind (datum condition)
      (ignore-errors (call-data-reader grid (record-data record) column)) 
    (cond (condition	nil)
          ((null datum)	nil)
          (t (values (funcall (column-data-read-converter column) datum)
                     (list :value datum))))))

(defmethod set-cell-clip ((grid ygrid) (record record) (column text-input-column-mixin)
                          (format (eql :string)) (text string)
                          &key (redraw-now :header))
 ;;; Used by: Edit/select
  (save-cell grid record column  :redraw-now redraw-now  :text text))
  ;(write-cell-text grid record column text :redraw-now nil))
 
(defmethod revert-cell ((grid ygrid) (record record) (column text-column))
  (let* ((pinboard (pinboard grid))
         (overlay (overlay pinboard)))
    (with-slots (%old-value%) grid	;text-input buffer
      (when (and (neq %old-value% *unbound-value*)
                 (eq (current-displayer pinboard) column)
                 (eq (current-subscript pinboard) record))
                 ;(eq (capi:capi-object-property text-input 'column) column))
        (when overlay
          (let ((text (funcall (column-data-read-converter column) %old-value%))
                (buffer (capi:editor-pane-buffer overlay)))
            (when (editor:variable-value-if-bound 'editor:highlight-active-region
                                                  :buffer buffer)
              (editor::set-highlight-buffer-region nil buffer));(capi:call-editor "hl off")
            (setf (capi:editor-pane-text overlay) text
                  (editor:buffer-modified buffer) nil)
            (capi:call-editor overlay "Clear Undo")) )
        ; let :after method redraw
        (when (column-data-writer column)
          (call-data-writer %old-value% grid (record-data record) column))
        ;(set-cell-value grid record column %old-value% :redraw-now nil)
) ) ) )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BUFFER  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The below commands must be in sync with the function record-board-on-gesture
;;; invoked when the editor pane is not visible.
;;; Q: Some stuff stop working after highlighting characters in the pane?

(defun buffer-empty-or-at-first-line-p ()
 ;;; Workaround: (editor::first-line-p (editor:current-point)) => nil if buffer is empty
  (or (editor::first-line-p (editor:current-point))
      (= (editor:buffer-character-count (editor:current-buffer)) 0)))

(editor:defcommand "Grid Left" (p) "" ""
  (declare (ignore p))
  (if (and (editor:start-line-p (editor:current-point))
           (buffer-empty-or-at-first-line-p))
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane))
                        0 0 nil :left)
      #+win32 (editor::backward-character-cancelling-selection-command nil)
      #-win32 (editor:backward-character-command nil) ))
(editor:defcommand "Grid Right" (p) "" ""
  (declare (ignore p))
  (if (and (editor:end-line-p (editor:current-point))
           (editor::last-line-p (editor:current-point)))
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane))
                        0 0 nil :right)
      #+win32 (editor::forward-character-cancelling-selection-command nil)
      #-win32 (editor:forward-character-command nil) ))

(editor:defcommand "Grid Backward Word" (p) "" ""
  (declare (ignore p))
  (if (and (editor:start-line-p (editor:current-point))
           (buffer-empty-or-at-first-line-p))
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane))
                        0 0 nil :left)
      #+win32 (editor::backward-word-cancelling-selection-command nil)
      #-win32 (editor:backward-word-command nil) ))
(editor:defcommand "Grid Forward Word" (p) "" ""
  (declare (ignore p))
  (if (and (editor:end-line-p (editor:current-point))
           (editor::last-line-p (editor:current-point)))
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane))
                        0 0 nil :right)
      #+win32 (editor::forward-word-cancelling-selection-command nil)
      #-win32 (editor:forward-word-command nil) ))

(editor:defcommand "Grid Tab" (p) "" ""
  (declare (ignore p))
  ;(yl:debug-format :grid "Grid Tab > ~S" (editor::current-editor-pane))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0
                    #.(sys::make-gesture-spec (char-code #\Tab) 0)))
(editor:defcommand "Grid ShiftTab" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0
              #.(sys::make-gesture-spec (char-code #\Tab) sys:gesture-spec-shift-bit)))

(editor:defcommand "Grid Home" (p) "" ""
  (declare (ignore p))
  (if (editor:start-line-p (editor:current-point))
      #+win32 (editor::beginning-of-buffer-cancelling-selection-command nil)
      #-win32 (editor:beginning-of-buffer-command nil)
      #+win32 (editor::beginning-of-line-cancelling-selection-command nil)
      #-win32 (editor:beginning-of-line-command nil)))
(editor:defcommand "Grid End" (p) "" ""
  (declare (ignore p))
  (if (editor:end-line-p (editor:current-point))
      #+win32 (editor::end-of-buffer-cancelling-selection-command nil)
      #-win32 (editor:end-of-buffer-command nil)
      #+win32 (editor::end-of-line-cancelling-selection-command nil)
      #-win32 (editor:end-of-line-command nil)))

(editor:defcommand "Grid Page Left" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane))
                    0 0 nil :page-left))
(editor:defcommand "Grid Page Right" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane))
                    0 0 nil :page-right))

(editor:defcommand "Grid Leftmost" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :home))
(editor:defcommand "Grid Rightmost" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :end))

(editor:defcommand "Grid Up" (p) "" ""
  (declare (ignore p))
  (if (buffer-empty-or-at-first-line-p)
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :up)
      #+win32 (editor::previous-line-cancelling-selection-command nil)
      #-win32 (editor:previous-line-command nil) ))
(editor:defcommand "Grid Down" (p) "" ""
  (declare (ignore p))
  (if (editor::last-line-p (editor:current-point))
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :down)
      #+win32 (editor::next-line-cancelling-selection-command nil)
      #-win32 (editor:next-line-command nil) ))

(editor:defcommand "Grid Page Up" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :page-up))
(editor:defcommand "Grid Page Down" (p) "" ""
  (declare (ignore p))
  (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :page-down))

(editor:defcommand "Grid First" (p) "" ""
  (declare (ignore p))
  (if (and (editor:start-line-p (editor:current-point))
           (buffer-empty-or-at-first-line-p))
      (pinboard-on-move (capi:element-parent(editor::current-editor-pane)) 0 0 nil :first)
      #+win32 (editor::backward-paragraph-cancelling-selection-command nil)
      #-win32 (editor:backward-paragraph-command nil) ))
(editor:defcommand "Grid Last" (p) "" ""
  (declare (ignore p))
  (if (and (editor:end-line-p (editor:current-point))
           (editor::last-line-p (editor:current-point)))
      (pinboard-on-move (capi:element-parent (editor::current-editor-pane)) 0 0 nil :last)
      #+win32 (editor::forward-paragraph-cancelling-selection-command nil)
      #-win32 (editor:forward-paragraph-command nil) ))


(editor:defcommand "Grid Delete Record" (p) "" ""
  (declare (ignore p))
  (record-board-on-gesture (capi:element-parent (editor::current-editor-pane)) 0 0
         #.(sys:make-gesture-spec (char-code #\-) sys:gesture-spec-control-bit)))

#|(editor:defcommand "Grid Cell Action" (p) "" ""
  (declare (ignore p))
  (cell-on-action (capi:element-interface (editor::current-editor-pane))))|#

(editor:defcommand "Grid Press Return" (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (case (capi:capi-object-property grid :press-return)
      (:action	 (cell-on-action grid))
      (:down	 (pinboard-on-move (pinboard grid) 0 0 nil :down))
      (otherwise (pinboard-on-move (pinboard grid) 0 0 #.(sys::make-gesture-spec
                                                          (char-code #\Tab) 0))))))

(editor:defcommand "Grid Save Record" (p) "" ""
  (declare (ignore p))
  (save-record (capi:element-interface (editor::current-editor-pane))))

(editor:defcommand "Grid Revert Cell"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (revert-cell grid (%current-record grid) (%current-column grid))))

(editor:defcommand "Grid Complete Cell"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (complete-cell grid (%current-record grid) (%current-column grid))))

(editor:defcommand "Grid Select All"  (p) "" ""
  (declare (ignore p))
  (select-section (capi:element-interface (editor::current-editor-pane)) :all))

(editor:defcommand "Grid Indent"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (operate-on-cell grid (%current-record grid) (%current-column grid) :indent)))

(editor:defcommand "Grid Unindent"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (operate-on-cell grid (%current-record grid) (%current-column grid) :unindent)))

(editor:defcommand "Grid Search"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (operate-on-grid (capi:top-level-interface grid) grid 'searcher)))
    ;(capi:find-interface 'searcher
    ;                   :grid (capi:element-interface (editor::current-editor-pane))))

(editor:defcommand "Grid Zoom"  (p) "" ""
  (declare (ignore p))
  (operate-on-focus (editor::current-editor-pane) :zoom))

(editor:defcommand "Grid Popup"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (operate-on-cell grid (%current-record grid) (%current-column grid) :popup)))

(editor:defcommand "Grid Quote"  (p) "" ""
  (declare (ignore p))
  (let* ((pane (editor::current-editor-pane))
         (buffer (capi:editor-pane-buffer pane))
         (grid (capi:element-interface pane))
         (current-row (current-row grid)))
    (if (and (editor:buffer-writable buffer) (iplusp current-row))
        (when-let (string (cell-clip grid
                                     (row-record grid (1- current-row))
                                     (current-column grid)))
          (setf (capi:editor-pane-text pane) string
                (editor:buffer-modified buffer) t)
          #+win32 (editor::end-of-buffer-cancelling-selection-command nil)
          #-win32 (editor:end-of-buffer-command nil))
        (beep))))

(editor:defcommand "Grid Undo"  (p) "" ""
  (let* ((pane (editor::current-editor-pane))
         (buffer (capi:editor-pane-buffer pane)))
    (if (editor:buffer-modified buffer) ;(and (editor:buffer-writable buffer))
        ;(not (ywi:undo-buffer-empty-p (capi:editor-pane-buffer pane)))
        (editor:undo-command p)
        (let ((grid (capi:element-interface pane)))
          (revert-cell grid (%current-record grid) (%current-column grid))
          (beep)))))

;;; Document-scoped commands, e.g. for file-interface
;;; callback-type is usually :interface-data

(editor:defcommand "Grid Refresh"  (p) "" ""
  (declare (ignore p))
  (let ((pane (editor::current-editor-pane)))
    (operate-on-grid (capi:top-level-interface pane) (capi:element-interface pane)
                     :refresh)))

(editor:defcommand "Grid Goto"  (p) "" ""
  (declare (ignore p))
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    (operate-on-grid (capi:top-level-interface grid) grid 'goto-record)))

(editor:defcommand "Grid Help"  (p) "" ""
  (declare (ignore p))
  (let ((pane (editor::current-editor-pane)))
    (unless (capi:help-key pane)			; -> beep if help-key is null
      (ywi:help-callback (capi:top-level-interface pane) pane :help nil))))

(editor:defcommand "Delete Character or Selection" (p)
     "Delete (not kill!) the region if it is highlighted otherwise - next character" ""
  (let* ((pane (editor::current-editor-pane))
         (buffer (capi:editor-pane-buffer pane)))
    (cond ((editor:variable-value-if-bound 'editor:highlight-active-region
                                           :buffer buffer)
           (editor:delete-region-command p)	; delete without pushing into the kill ring
           (editor::set-highlight-buffer-region nil buffer)
           ;(editor:"Delete Selection Mode")   
          )
          ((editor:delete-next-character-command p)))))

#|;; Deprecated in favour of  "Save", "New, and "Print"
(editor:defcommand "Grid New"  (p) "" ""
  (declare (ignore p))
  (operate-on-interface (capi:top-level-interface (editor::current-editor-pane)) :new))

(editor:defcommand "Grid Save"  (p) "" ""
  (declare (ignore p))
  (operate-on-interface (capi:top-level-interface (editor::current-editor-pane)) :save))

(editor:defcommand "Grid Print"  (p) "" ""
  ;; CAUTION: For rr:menu-run callback, callback-type is :full.
  ;;          Can grid be embedded into report-board?
  (declare (ignore p))
  (operate-on-interface (capi:top-level-interface (editor::current-editor-pane)) :print))|#

;;; Locale-specific character case convertion

(defun locale-nstring-upcase$simple-string (string start end)
 ;;; A la system::nstring-upcase$simple-string
  (declare (type simple-string string) (fixnum start end))
  (do ((i start (1+ i)))
      ((>= i end) string)
    (declare (fixnum i))
    (setf (schar string i) (locale-char-upcase (schar string i)))))

(defun locale-nstring-downcase$simple-string (string start end)
 ;;; A la system::nstring-downcase$simple-string
  (declare (type simple-string string) (fixnum start end))
  (do ((i start (1+ i)))
      ((>= i end) string)
    (declare (fixnum i))
    (setf (schar string i) (locale-char-downcase (schar string i)))))

(defmacro with-point-and-mark (&body body)
  `(let ((%point% (editor::current-point))
	 (%mark% (editor::current-mark)))
     (if (editor::point> %point% %mark%)
	 (rotatef %point% %mark%))
     ,@body))

(editor:defcommand "Uppercase Word" (p)
  "Uppercase a word at point. With prefix argument uppercase that many words."
  "Uppercase p words at the point."
  (editor::filter-words p (editor::current-point) 'locale-nstring-upcase$simple-string))

(editor:defcommand "Lowercase Word" (P)
  "Lowercase a word at point. With prefix argument lowercase that many words."
  "Lowercase p words at the point."
  (editor::filter-words p (editor::current-point) 'locale-nstring-downcase$simple-string))

(editor:defcommand "Uppercase Region" (p)
  #1="Uppercase words from point to mark." #1#
  (declare (ignore p))
  (with-point-and-mark
   (editor::change-region-case %point% %mark% 'locale-nstring-upcase$simple-string)))

(editor:defcommand "Lowercase Region" (P)
  #1="Lowercase words from point to mark." #1#
  (declare (ignore p))
  (with-point-and-mark 
   (editor::change-region-case %point% %mark% 'locale-nstring-downcase$simple-string)))


(editor:defmode "Grid"
  :setup-function 'setup-grid-mode
  :key-bindings
  '(("Grid Left"		#\Left)
    ("Grid Right"		#\Right)
    ("Grid Tab"			#\Tab)
    ("Grid ShiftTab"		#\Shift-Tab)
    ("Grid Backward Word"	#\Control-Left)
    ("Grid Forward Word"	#\Control-Right)
    ("Grid Home"		#\Home)
    ("Grid End"			#\End)
    ;("Grid Page Left"		#\Meta-Prior)	; :emacs only
    ;("Grid Page Right"		#\Meta-Next)	; :emacs only
    ("Grid Leftmost"		#\Control-Home)
    ("Grid Rightmost"		#\Control-End)

    ("Grid Up"			#\Up)
    ("Grid Down"		#\Down)
    ("Grid Page Up"		#\Prior)
    ("Grid Page Down"		#\Next)
    ("Grid First"		#\Control-Up)		; first record
    ("Grid Last"		#\Control-Down)		; last record

    ("Prior Tab Page"		#\Control-Prior)
    ("Next Tab Page"		#\Control-Next)
    ("Prior Frame"		#\Shift-F6)
    ("Next Frame"		#\F6)

    ("Grid Delete Record"	#\Control--)

    ("New Line"			#\Control-Return)
    ("Grid Save Record"		#\Shift-Return)
    
    ("Grid Press Return"	#\Return)

    ;("Undo" #\Meta-Backspace)				; :emacs only
    ("Grid Undo"		#\Control-\z)
    ("Grid Undo"		#\Control-\)
    ("Grid Revert Cell"		#\Control-\Z)		; =Ctrl+Shift+Z
    #+Russian ("Grid Revert Cell" #\Control-\)		; =Ctrl+Shift+
    ("Grid Complete Cell"	 #.*completion-gesture*) ; #\Control-\Space, a la Access
    ("Grid Select All"		 #\Control-\a)
    #+Russian ("Grid Select All" #\Control-\)
    ("Mark Whole Buffer"	 #\Control-\A)
    #+Russian ("Mark Whole Buffer" #\Control-\)

    ("Kill Region"		#\Control-\x)
    #+Russian ("Kill Region"	#\Control-\)
    ("Un-Kill"			#\Control-\v)
    #+Russian ("Un-Kill"	#\Control-\)
    ("Save Region Preserving Selection"           #\Control-c)
    #+Russian ("Save Region Preserving Selection" #\Control-)

    ("Grid Search"		#\Control-\f)
    #+Russian ("Grid Search"	#\Control-\)
    ("Grid Goto"		#\Control-\g)
    #+Russian ("Grid Goto"	#\Control-\)
    ("Grid Quote"		#\Control-\')		; does not work on some OSes
    #+Russian ("Grid Quote"	#\Control-\)
    ;("Grid Quote"		#\Control-\q)
    ;#+Russian ("Grid Quote"	#\Control-\)

    ("Grid Indent"		#\Control-\j)		; Indent New Line
    #+Russian  ("Grid Indent"	#\Control-\)		; Indent New Line
    ("Grid Unindent"		#\Control-Shift-\j)
    #+Russian  ("Grid Unindent"	#\Control-Shift-\)
    ("Do Nothing"		#\Meta-\j)		; Indent New Comment Line

    ("Grid Zoom"		#\Shift-F2)
    ("Grid Popup"		#\F4)
    ;("Grid Popup"		"Meta-Down")		; :emacs only
    ("Grid Refresh"		#\F5)			; Refresh screen

    ("New"			#\Control-\n)		; New buffer
    #+Russian ("New"		#\Control-\)
    ("Open"			#\Control-\o)		; Indent New Line
    #+Russian ("Open"		#\Control-\)
    ("Print"			#\Control-\p)		; Print file
    #+Russian ("Print"		#\Control-\)
    ("Save"			#\Control-\s)		; Incremental search
    #+Russian ("Save"		#\Control-\)

    ;; Redefine to Nothing commands requiring Echo area and other dangerous.
    ("Grid Help"  #\F1)			; Let invoke context help if any
    ("Do Nothing" #\Control-F4)		; Kill buffer - TODO: Close window in MDI?
    ("Do Nothing" #\Control-F6)		; Switch to next/previous window
    ("Do Nothing" #\Control-Shift-F6)
    ("Do Nothing" #\Control-F11)	; Describe System
    ("Do Nothing" #\Control-\r)		; Reverse incremental search
    ;("Do Nothing" #\Control-\.)	; Expand File Name

    ;; Platform-specific binding?
    ("Delete Character or Selection" #\Rubout)	; key-binds.lisp in lww-ide.zip

    ;; Case change uses Control instead of Meta
    ("Uppercase Word"		#\Control-\u)		; #\meta-\u
    #+Russian ("Uppercase Word"	#\Control-\)
    ("Lowercase Word"		#\Control-\l)		; #\meta-\l
    #+Russian ("Lowercase Word"	#\Control-\)
    ("Uppercase Region"		#\Control-\U)		; #\meta-\U
    #+Russian ("Uppercase Region" #\Control-\)
    ("Lowercase Region"		#\Control-\L)		; #\meta-\L
    #+Russian ("Lowercase Region" #\Control-\)
)  )

(defun setup-grid-mode (buffer)
  (declare (ignore buffer)))
  ;(setf (editor:buffer-major-mode buffer) 
  ;      (if (editor::get-mode-object "Pc Echo" t) "Pc Echo" "Echo Area"))

(defparameter *grid-text-input-model*
 ;;; NB: For invoking post-menu via SHIFT+F10, the input model does not help
  ;;     and the menu bar is activated instead. Nothing works:
  ;;     - neither (:keyboard-post-menu pane-post-menu) 
  ;;     - nor (editor:defcommand "Grid Post Menu" ...).
  ;;     Only explicit accelerator could help.
  (mapcar (lambda (pair &aux (first (first pair)))
            (cond ((equal first '(:button-1 :press))
                   (list first (second pair) 'editor-press))
                  ((equal first '(:button-1 :second-press))
                   (list first (second pair) 'editor-second-press))
                  (pair)))
          CAPI::*EDITOR-INPUT-MODEL*))

 #|(sublis '((capi::editor-start-region . editor-press)
             (capi::editor-start-mark-form . editor-second-press))
           capi::*editor-input-model*))|#

(defun editor-interrupt-hook ()
 ;;; React on #\Escape.
  ;; Editor interrupt hook is a list of functions and arguments 
  ;; evaluated by means of sys::call-functions-list.
  (let ((grid (capi:element-interface (editor::current-editor-pane))))
    ;(capi:execute-with-interface grid #'
    (cell-on-cancel grid (%current-column grid))))

(defun initialize-editor-buffer (pane buffer)
  (setf (editor:buffer-minor-mode buffer "Grid") t
        (capi:output-pane-input-model pane) *grid-text-input-model*
        (editor:variable-value 'editor::editor-interrupt-hook :buffer buffer)
         '(editor-interrupt-hook)))
 ;(cons (list 'mp:kill-background-job (mp:process-current-job mp:*current-process*))
 ;      (delete 'mp:kill-background-job 
 ;		 (variable-value-if-bound 'editor-interrupt-hook :buffer buffer)
 ;		 :key 'car)))))
