;;;-*- Mode: Lisp; -*-
;;; YstokGrid - LispWorks CAPI-based grid widget
;;; Copyright (c) 2003-2013 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Externals

(in-package :ystok.grid)

(defgeneric add-record (grid data &key count redraw-now &allow-other-keys)
 (:method ((grid ygrid) data &rest args &key count
                                             (redraw-now (CAPI::PANE-CREATED-P grid))
                                             selected)
 "Create a new record and put at the last or last but one position.
Arguments:
 DATA : Data object
 COUNT: If true, increase the allow-additions value.
 SELECTED: If true, select the record added.
 ARGS : Plist-like list of arguments passed to GRID's record consturctor.
 Value: New record"
  (with-slots (nrows records selected-rows) grid
    ;(assert (eql ncols (length data)))
    (let (record record-row-number auto-record)
      (without-properties (args args :count :redraw-now)
        (if (setq auto-record (record-auto-p grid (1- nrows)))
          ;; Insert a record for the new data to the last but one position and
          ;; move the auto-record to the last position
          (progn
            (vector-push-extend auto-record records *records-increment*)
            (setf record (apply (slot-value grid 'record-constructor)
                                :row-number (setq record-row-number (1- nrows))
                                :data data
                                args)
                  (aref records (1- nrows)) record
                  (record-row-number auto-record) nrows))
          ;; Else simply create a new record and push to the end
          (vector-push-extend (setq record (apply (slot-value grid 'record-constructor)
                                                :row-number (setq record-row-number nrows)
                                                :data data
                                                args))
                              records *records-increment*)))
      (incf nrows)
      ;(assert (eql nrows (fill-pointer records)))
      (when selected
        (setf (current-row grid) record-row-number)
        (push record-row-number selected-rows))
      (when (and count (integerp (allow-additions grid)))	; as addition can be
        (incf (slot-value grid 'allow-additions)))		; called programmaticaly
      (when redraw-now
        (unless (%reset-bars grid :horizontal nil)	; redraw has not called internally
          (invalidate-row grid record-row-number :redraw-now :all)
          (when auto-record
            (invalidate-record grid auto-record :redraw-now :all))))
      record))) )

(defgeneric (setf allow-additions) (value grid)
 (:method (value (grid ygrid))
  (let ((old-value (allow-additions grid)))
    (cond ((and value (not old-value))
           (setf (slot-value grid 'allow-additions) (if (integerp value) value 0))
           (when (CAPI::PANE-CREATED-P grid)
             (%fill-auto-record grid nil :redraw-now t)))
          ((and (not value) old-value)
           (setf (slot-value grid 'allow-additions) nil)
           (when (and (CAPI::PANE-CREATED-P grid) (record-auto-p grid (1- (nrows grid))))
             (vector-pop (records grid))
             (when (and (eql (decf (nrows grid)) (current-row grid))
                        (plusp (current-row grid)))
               (decf (current-row grid)))
             (%invalidate-grid grid))) ))) )

(defun nilf-from-selection (row-to-delete selected-rows)
 ;;; Replace row-to-delete by nil and decrease greater row numbers by 1
  (declare (optimize (speed 3) (safety 0)))
  (loop for cons on selected-rows
        for row = (car cons)
        when row 
        if (= row row-to-delete) do (rplaca cons nil)
        else if (> row row-to-delete) do (rplaca cons (1- row)))
  selected-rows)

(defgeneric delete-record (grid &key interactive)
 (:documentation "Delete the record or row specified.
Arguments:
 ROW or RECORD:	Is not obliged to be current.
 INTERACTIVE :	If eql :recursive, the record-data has been deleted already,
  	 	and we are only excluding the record from the records vector.
  		In this case, record-on-delete is not invoked but
		delete-callback is called (recalculating totals etc.)
 CURRENT :	Try invoke current-callback if the current record was deleted.
Value: True   For auto-record or recursive or the one returned by record-on-delete.
       False  Usually indicates a failure.")
  ;; Q: Do we need to put the grid-confirm-delete call inside the handler?
 (:method ((grid ygrid) &key record
                             (row (if record (record-row-number record)(current-row grid)))
                             (interactive t) (redraw-now t) (current t))
  (with-slots (records selected-rows nrows current-row) grid
   (cond ((< row nrows)
          (let ((record (or record (%row-record grid row)))
                (current-record (current-record grid))		; old current
                value)
            (cond ((setq value (record-auto-p grid row))	; don't delete auto record!
                   (setf (record-selected record) nil
                         selected-rows (nilf-from-selection row selected-rows)))
                  ((setq value
                         (or (eq interactive :recursive)
                             (getf (record-plist record) :new)
                             (if (grid-confirm-delete grid record interactive)
                                 (restart-case
                                     (handler-bind ((error #'prompt-for-restart))
                                       (record-on-delete grid record))
                                   (abort ()
                                    :report (list #L"(Cancel) Abort deletion of row ~d."
                                                  (1+ row))
                                    (return-from delete-record nil)) )
                                 nil)))
                   (when (= row current-row)
                     (%set-edit-mode grid nil nil :redraw-now nil))
                   (setf selected-rows (nilf-from-selection row selected-rows))
                   (when (< row (1- nrows))              ; no replace if record is the last
                     (replace records records :start1 row :start2 (1+ row)))
                   (when (and (>= current-row row) (> current-row 0))
                     (decf current-row))
                   (vector-pop records)
                   (decf nrows)
                   (loop for r upfrom row below nrows
                         do (setf (record-row-number (%row-record grid r)) r))
                   ;; DI 2007-Apr-17: Code snippet moved from record-on-delete
                   (when-let (delete-callback (slot-value grid 'delete-callback))
                     (funcall delete-callback grid record)) ))
            (when value
              (when interactive
                (setf selected-rows (delete nil selected-rows))
                (when redraw-now
                  (%invalidate-grid grid t)))
              (when current
                (%record-on-current grid current-record)))
            value))
         (interactive
          (beep))))) )

(defun delete-selected-rows (grid)
 ;;; We must always start deleting from the largest row
  ;; Values: List of values returned by delete-record in reverse order
  ;; NB: The specialization on #<grid-confirm-delete ... (eql :rows)> can change
  ;;     *confirm-deletions* that is later observed by the specialization on
  ;;	 #<grid-confirm-delete ... record>
  (with-slots (selected-rows) grid
    (when selected-rows
      (let ((*confirm-deletions* *confirm-deletions*)
            (current-record (current-record grid))
            (acc ()))
        (when (grid-confirm-delete grid :rows (length selected-rows))
          ;(setf selected-rows (sort selected-rows #'<))
          (with-bulk (#L"(Cancel) Abort bulk deletion.")
            (do-selected-records (record grid)
              (setq acc (cons (delete-record grid :record record
                                             :interactive nil :redraw-now nil :current nil)
                              acc))))
          (when (setf selected-rows (delete nil selected-rows))
            ;; As some records remain selected, ensure current row to be inside
            (unless (member (current-row grid) selected-rows)
              (setf (current-row grid) (first selected-rows))))
          (%invalidate-grid grid t)
          (%record-on-current grid current-record)
          acc)))))

(defun delete-record-or-rows (grid)
  (if (selected-rows grid)
      (delete-selected-rows grid)
      (delete-record grid)))

(defun goto-record (grid &key data record row)
 ;;; Q: Add parameter enter?
  (when (orf row (and (orf record (find-record grid data))
                      (record-row-number record)))
    (unless (eql row (current-row grid))
      (pinboard-on-move (pinboard grid) 0 0 row :record-number))
      ; grid row))
    row))

(defun goto-column (grid &key column col)
  (let ((column (find-column grid (or column col))))
    (if (and column
             (column-enabled grid column)
             (not (column-hidden column)))
        (progn (pinboard-on-move (pinboard grid) 0 0
                                 (setq col (column-number column)) :column-number)
          col)
        (beep))))

(defun goto (grid record column)
 ;;; Value: True <= pinboard-on-move, current cell has been actually shifted.
  ;;        False Movement is not needed (or error?)
  ;; Q: Add parameter enter?
  (let* ((row (cond ((integerp record) record)
                    ((record-p record) (record-row-number record))))
         (col (when (and column
                         (setq column (find-column grid column))
                         (column-enabled grid column)
                         (not (column-hidden column)))
                (column-number column)))
         (pinboard (pinboard grid)))
    (when (eql row (current-row grid))
      (setq row nil))
    (when (eql col (current-col grid))
      (setq col nil))
    (cond ((and row col)
           (pinboard-on-move pinboard 0 0 (cons row col) :pan))
          (col
           (pinboard-on-move pinboard 0 0 col :column-number))
          (row
           (pinboard-on-move pinboard 0 0 row :record-number))
          ((beep nil pinboard)))))

(defgeneric new-record (grid)
 (:documentation "Move the current pointer to the empty row at the grid bottom")
 (:method ((grid basic-grid))
  (pinboard-on-move (pinboard grid) nil nil nil :new)) ) ;(capi:execute-with-interface grid

;(defmethod operate-on-grid-p (interface (grid ygrid) (operation (eql 'save-record)))
;  (when-let (record (current-record grid))
;    (< (record-row-number record) (nrows grid))))
       ;(%get-edit-mode grid))))			; we may not leave the cell now

(defgeneric find-record (grid data &key test &allow-other-keys)
 (:method ((grid ygrid) data &key (test #'eq))
  (find data (records grid) :key #'record-data :test test)))

(defgeneric find-column (grid column-designator &key errorp test)
 ;;; Args: test Function of signature: (column-designator column-name),
  ;;            i.e column-designator is passed first!
 (:method (grid (column basic-column) &key errorp test)
  (declare (ignore grid errorp test))
  column)

 (:method (grid (col integer) &key errorp test)
  (declare (ignore errorp test))
  (cond ((cond (test
                (find col (columns grid) :key #'column-name :test test))
               ((< -1 col (length (columns grid)))
                (col-column grid col))))
        (errorp (error #L"Column ~s is not found in grid~@[ '~a'~]."
                       col (object-title grid)))))

 (:method (grid column-name &key (errorp t) (test #'equalp))
  (cond ((find column-name (columns grid) :key #'column-name :test test))
        (errorp (error #L"Column ~s is not found in grid~@[ '~a'~]."
                       column-name (object-title grid))))))

(defun cell-value (grid &optional (row-or-record (current-row grid))
                                  (column-designator (current-col grid)))
  (when-let (record (if (integerp row-or-record)
                        (row-record grid row-or-record)
                        row-or-record))
    (call-data-reader grid (record-data record) (find-column grid column-designator
                                                             :errorp t))))

(defun (setf cell-value) (value grid &optional (row-or-record (current-row grid))
                                               (column-designator (current-col grid)))
 ;;; Call set-cell-value directly instead of save-cell as checking nullable
  ;; and handling conditions seems not needed.
  (set-cell-value grid row-or-record column-designator value)
  value)

(defgeneric cell-old-value (grid row-or-record column)
  (:method ((grid ygrid) (record record) (column basic-column))
   (if-bind (pair (assoc column (changed record)))
     (cdr pair)		; (if (eq *unbound-value*))?
     (call-data-reader grid (record-data record) column)))

  (:method ((grid ygrid) (record record) column-designator)
   (cell-old-value grid record (find-column grid column-designator :errorp t))) )

(defun (setf left-frozen-ncols) (value grid)
  (let ((value (max (min value (ncols grid)) 0)))
    (setf (slot-value grid 'left-frozen-width) nil
          (slot-value grid 'left-frozen-ncols) value)
    (pinboard-redraw grid)
    (column-header-redraw grid)
    ;grid (max (min value (ncols grid)) 0))
    value))

(defgeneric (setf column-width) (width grid column)
  (:method ((width integer) (grid ygrid) (column column))
   (%set-column-width width grid column))

  (:method ((width number) (grid ygrid) col)
   (setf (column-width grid (find-column grid col :errorp t)) (round width)))

  (:method ((width (eql :best-width)) (grid ygrid) (column column))
  ;; Maximum over the width of all records content or the heading width.
   (let ((size (compute-natural-size grid column))
         (title (when (iplusp (gp:port-height (column-header grid)))
                  (%initial-column-pixels grid column '(:string :title)))))
     (when (or size title)
       (setf (column-width grid column) (or (and size title (max size title))
                                            size title))))))

(defun (setf row-height) (height grid)
  (%set-row-height height grid))

(defun (setf gridlines) (value grid)
  (with-slots (gridlines) grid
    (setf gridlines value)
    (pinboard-redraw grid)))
    ;(gp:invalidate-rectangle pinboard)))

;;; Enabled ::= true	   Normal
;;;		:read-only Does not allow edits, but allow enter and copy data.
;;;		nil	   Disabled, i.e. does not even allow to tab or click into column
(defgeneric column-enabled (grid column)
 (:method ((grid ygrid) (column column))
  (declare (ignore grid))
  (enabled column))
 (:method ((grid ygrid) col)
   (column-enabled grid (find-column grid col :errorp t))) )

(defgeneric (setf column-enabled) (value grid column &optional data)
 ;;; Args: data If true, also invalidate the data cache.
 (:method (value (grid ygrid) (column column) &optional data)
  (if (setf (slot-value column 'enabled) value)
      (remf (column-graphics-args column) :foreground)
      (setf (getf (column-graphics-args column) :foreground)
            #-(and lispworks4 win32) :color_graytext
            #+(and lispworks4 win32) 'win32:color_graytext))
  (invalidate-column grid column :data data :redraw-now t)
  value)
 (:method (value (grid ygrid) col &optional data)
  (setf (column-enabled grid (find-column grid col :errorp t) data) value)) )

(defun unhide-column (grid column &optional width)
 ;;; Restore the column width to given or default extracted from the column-description
  (when (eql (column-width (setq column (find-column grid column))) 0)
    (setf (column-width grid column) (%initial-column-pixels grid column width))))

(defun show-enable-column (grid column value &key width redraw-now)
 ;;; Unhide/hide and enable/disable simultaneously.
  ;; Args: width Default width of the column in shown mode.
  (setf column (find-column grid column :errorp t)
        (column-enabled grid column) (if value t nil))		; to boolean
  (if value
      (unless width
        (setq width (%initial-column-pixels grid column)))
      (setq width 0))
  (when (xor (iplusp (column-width column)) (iplusp width))
    (if redraw-now
        (setf (column-width grid column) width)
        (progn (setf (slot-value column 'width) width
                     (slot-value grid 'internal-width) nil)
          (when (< (column-number column) (left-frozen-ncols grid))
            (setf (slot-value grid 'left-frozen-width) nil))))))


(defmacro with-columns (column-descriptions grid &body body)
 ;;; Args: column-description ::= column-name		  - symbol (not evaluated)
  ;;                            | (var column-designator) - expression (evaluated)
  (lw:rebinding (grid)
   `(let ,(mapcar (lambda (description)
                    (let* ((variable (first-or-self description))
                           (value (if (consp description)
                                      (second description)
                                      `',variable)))
                      `(,variable (find-column ,grid ,value))))
                  column-descriptions)
      ,@body)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  OPERATIONS on GRID  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric fetch-grid (grid source &key &allow-other-keys)
 ;;; Args: count   If true, we are in "piecemeal mode", i.e. inserting records 
  ;;		   one-by-one and counting every addition.
  ;;       current If true, try to restore current column to current-col as it was before.
 (:method (grid (source list) &key count (redraw-now 30) plist limit)
  ;; Can also be called in a process not coinciding with grid's.
  (do (record
       (limit (the fixnum (or limit most-positive-fixnum)))
       (list source (rest list))
       (row 0 (1+ row)))
      ((or (null list) (>= row limit))
       row)
    (declare (fixnum limit row))
    (setf record (add-record grid (first list)
                             :count count
                             :redraw-now (cond ((numberp redraw-now)
                                                (zerop (mod row redraw-now)))
                                               ((eq redraw-now :after)
                                                nil)
                                               (redraw-now))
                             ;:selected count
                             :plist (if plist (copy-list plist) nil)))))

 (:method :after ((grid ygrid) source &key count (redraw-now :after) (focus t) current)
  (declare (ignore source))
  (unless count							; if in piecemeal mode,
    (setf (changed grid) nil))					; do not reset counters
  (when (and (allow-additions grid) (not (record-auto-p grid (1- (nrows grid)))))
    (%fill-auto-record grid nil :redraw-now nil))
  (when redraw-now
    #+lispworks4.3 (interface-on-create grid)
    (%invalidate-grid grid)
    (when (> (nrows grid) 0)
      (let ((record (%current-record grid)))
        ;; As we keep staying on the same row, to provide full-fledged move-cell,
        ;; call record-on-enter but do not enter auto-record!
        (%record-on-current grid nil record)
        (cond (focus
               (focus-grid grid))				; -> cell-on-enter anyhow
              ((and current (> (ncols grid) 0))			; verify current column
               (let ((column (if (< (current-col grid) (ncols grid))
                                 (%current-column grid))))
                 (cond ((and column (enabled column) (> (column-width column) 1))
                        (let ((pinboard (pinboard grid)))
                          (setf (current-displayer pinboard) column
                                (current-subscript pinboard) record)))
                       ((setq column (find-if (lambda (column)
                                                (and (enabled column)
                                                     (> (column-width column) 1)))
                                              (columns grid)))
                        (move-cell grid (record-row-number record)	; -> cell-on-enter
                                    (column-number column))))))))))) )
    ;grid redraw-now focus current)) )

(defmethod capi:set-pane-focus :after ((pinboard record-board) &key force-initial)
 ;;; Ensure entering the current cell, if any, on (re)loading grid the first time.
  ;; Args: force-initial If true, enter the cell no matter what pinboard previous state is.
  ;; NB: We are trying to preserve current row and column if possible.
  ;; CAPI NB pane-got-focus:
  ;;  It can be valueable for creating tab pages on calling capi:set-pane-focus from
  ;;  user's code. "Programmatically" means "on every call", including from CAPI events.
  (when (or force-initial (not (capi:capi-object-property pinboard 'focus-grid)))
    (setf (capi:capi-object-property pinboard 'focus-grid) t)
    (let* ((grid (capi:element-interface pinboard))
           (ncols (ncols grid))
           (nrows (nrows grid)))
      (when (and (> (ncols grid) 0) (> (nrows grid) 0))
        (let ((column (if (< (current-col grid) ncols) (%current-column grid)))
              (row (if (< (current-row grid) nrows) (current-row grid) 0)))
          (when (or (and column (enabled column) (> (column-width column) 1))
                    (setq column (find-if (lambda (column)
                                            (and (enabled column)
                                                 (> (column-width column) 1)))
                                          (columns grid))))
            (move-cell grid row (column-number column))))))))	;->cell-on-enter, Q: focus?

(defgeneric focus-grid (grid)
 ;;; Called just after (re)loading grid content
 (:method ((grid ygrid))
  (capi:set-pane-focus (pinboard grid) :force-initial t)))

(defun invalidate-grid (grid &rest args
                        &key (data t) (graphics t) (redraw-now :cell) column (filter :on)
                        &allow-other-keys)
 ;;; Args: args	   Can contain additional args for invalidate-column, e.g. title or items.
  ;;	   column  Column, column designator (name or number), or list to invalidate;
  ;;		   if null, invalidate entire grid.
  ;;       redraw-now  When true, seems to switch focus.
  ;;       filter ::= :on | :off | :all
  ;; NB: What about (records-off grid)?
  (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (when (or data graphics)
    (let (;(records (records grid))
          (args (sans args :data :graphics :redraw-now :column)))
      (if (consp column)
          (let ((columns (mapcar (lambda(column)(column-number (find-column grid column)))
                                 column)))
            (do-records (record grid :filter filter)
              (dolist (col columns)
                (%invalidate-record-cache record col graphics data)))
            (when args
              (dolist (col columns)
                (apply #'invalidate-column grid col
                       :data nil :graphics nil :redraw-now nil
                       args))))
          (let ((col (if (or (null column) (numberp column))
                         column
                         (column-number (find-column grid column)))))
            (do-records (record grid :filter filter)
              (%invalidate-record-cache record col graphics data))
            (when args
              (map nil (lambda (column) (apply #'invalidate-column grid column
                                               :data nil :graphics nil :redraw-now nil
                                               args))
                   (if column (list column) (columns grid))))) )))
  (nilf (slot-value grid 'internal-width) (slot-value grid 'left-frozen-width))
  (when redraw-now
    (%invalidate-grid grid (eq redraw-now :cell))))

(defgeneric load-grid (grid source &key &allow-other-keys))

(defmethod dsum (column-designator (grid ygrid) &key where default (filter :on) new)
  ;; Args: new   If true, count the new record as well;
  ;;		 usually should receive T when called from update/delete callback.
  #-debug (declare (optimize (speed 3) (safety 0) (debug 0)))
  (let (data datum
        (column (find-column grid column-designator)))
    (do-records (record grid :filter filter)	; :on = only shown records
      (setq data (record-data record))
      (when (setq datum (if (and (or new (not (getf (record-plist record) :new)))
                                 (imp where (funcall where data)))
                            (call-data-reader grid data column)))
        (if (numberp default)
            (incf default datum)
            (setq default datum))))
    default))
